#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Classes and functions for working with coordinates systems."""

__all__ = [
    "unitTypes",
    "Vector",
    "Position",
    "Size",
    "Vertices"
]

import numpy as np
from .tools import monitorunittools as tools

# list of applicable units
unitTypes = [
    None,
    '',
    'pix',
    'deg',
    'degFlat',
    'degFlatPos',
    'cm',
    'pt',
    'norm',
    'height'
]

# anchor offsets and names
_anchorAliases = {
    'top': -0.5,
    'bottom': 0.5,
    'left': 0.5,
    'right': -0.5,
    'center': 0
}


class Vector:
    """Class representing a vector.

    A vector is a mathematical construct that specifies a length (or magnitude)
    and direction within a given coordinate system. This class provides methods
    to manipulate vectors and convert them between coordinates systems.

    This class may be used to assist in positioning stimuli on a screen.

    Parameters
    ----------
    value : ArrayLike
        Array of vector lengths along each dimension of the space the vector is
        within. Vectors are specified as either 1xN for single vectors, and
        Nx2 or Nx3 for multiple vectors.
    units : str or None
        Units which `value` has been specified in. Applicable values are
        `'pix'`, `'deg'`, `'degFlat'`, `'degFlatPos'`, `'cm'`, `'pt'`, `'norm'`,
        `'height'`, or `None`.
    win : `~psychopy.visual.Window` or None
        Window associated with this vector. This value must be specified if you
        wish to map vectors to coordinate systems that require additional
        information about the monitor the window is being displayed on.

    Examples
    --------
    Create a new vector object using coordinates specified in pixel (`'pix'`)
    units::

        my_vector = Vector([256, 256], 'pix')

    Multiple vectors may be specified by supplying a list of vectors::

        my_vector = Vector([[256, 256], [640, 480]], 'pix')

    Operators can be used to compare the magnitudes of vectors::

        mag_is_same = vec1 == vec2  # same magnitude
        mag_is_greater = vec1 > vec2  # one greater than the other

    1xN vectors return a boolean value while Nx2 or Nx3 arrays return N-length
    arrays of boolean values.

    """
    def __init__(self, value, units, win):
        # Create a dict to cache values on access
        self._cache = {}
        # Assume invalid until validation happens
        self.valid = False

        # define some names used by `set`
        self.win = None
        self._requested = None
        self._requestedUnits = None

        self.set(value, units, win)

    def set(self, value, units, win=None):
        # Check inputs
        if win is None:
            win = self.win

        self.win = win  # set extras

        # If input is a Vector object, duplicate all settings
        if isinstance(value, Vector):
            self._requested = value._requested
            self._requestedUnits = value._requestedUnits
            self.valid = value.valid
            self.pix = value.pix
            if win is None:
                self.win = value.win

            return

        # Validate
        value, units = self.validate(value, units)

        # Set values
        self._requested = value
        self._requestedUnits = units
        setattr(self, self._requestedUnits, self._requested)

    def validate(self, value, units):
        """Validate input values.

        Ensures the values are in the correct format.

        Returns
        -------
        tuple
            Parameters `value` and `units`.

        """
        # Assume valid until shown otherwise
        self.valid = True

        # Check units are valid
        if units not in unitTypes:
            raise ValueError(
                f"Unit type '{units}' not recognised, must be one of: "
                f"{unitTypes}")

        # Get window units if units are None
        if units in (None, ''):
            units = self.win.units

        # Coerce value to a numpy array of floats
        try:
            value = np.array(value, dtype=float)
        except ValueError as err:
            self.valid = False
            raise err

        # Make sure each value is no more than 3D
        if len(value.shape) == 0:
            value = np.array([value, value])
            self.valid = True
        elif len(value.shape) == 1:
            self.valid = value.shape[0] <= 3
        elif len(value.shape) == 2:
            self.valid = value.shape[1] <= 3
            if value.shape[0] == 1:
                # Remove extraneous layer if single value
                value = value[0]
        else:
            self.valid = False

        # Replace None with the matching window dimension
        if (value == None).any() or np.isnan(value).any():
            win = Vector((1, 1), units="norm", win=self.win)
            if len(value.shape) == 1:
                value[value == None] = getattr(win, units)[value == None]
                value[np.isnan(value)] = getattr(win, units)[np.isnan(value)]
            else:
                value[np.isnan(value[:, 0]), 0] = getattr(win, units)[0]
                value[np.isnan(value[:, 1]), 1] = getattr(win, units)[1]
                value[value[:, 0] == None, 0] = getattr(win, units)[0]
                value[value[:, 1] == None, 1] = getattr(win, units)[1]

        assert self.valid, (f"Array of position/size values must be either "
                            f"Nx1, Nx2 or Nx3, not {value.shape}")

        return value, units

    def __bool__(self):
        return self.valid

    def __repr__(self):
        """If vector is printed, it will display its class and value."""
        if self:
            return (f"<psychopy.layout.{self.__class__.__name__}: "
                    f"{np.round(self.pix, 3)}px>")
        else:
            return "<psychopy.layout.{self.__class__.__name__}: Invalid>"

    # --------------------------------------------------------------------------
    # Rich comparisons
    #

    def __eq__(self, target):
        """`==` will compare position in pix"""
        if isinstance(target, Vector):
            if self.pix.size > 1:
                return all(self.pix == target.pix)
            else:
                return self.pix == target.pix
        else:
            return False

    def __ne__(self, target):
        """`!=` will return the opposite of `==`"""
        return not self == target

    def __lt__(self, target):
        """`<` will compare magnitude"""
        if isinstance(target, Vector):
            return self.magnitude < target.magnitude
        elif isinstance(target, (int, float)):
            return self.magnitude < target
        else:
            return False

    def __le__(self, target):
        """`<=` will compare magnitude"""
        if isinstance(target, Vector):
            return self.magnitude <= target.magnitude
        elif isinstance(target, (int, float)):
            return self.magnitude <= target
        else:
            return False

    def __gt__(self, target):
        """`>` will compare magnitude"""
        if isinstance(target, Vector):
            return self.magnitude > target.magnitude
        elif isinstance(target, (int, float)):
            return self.magnitude > target
        else:
            return False

    def __ge__(self, target):
        """`>=` will compare magnitude"""
        if isinstance(target, Vector):
            return self.magnitude >= target.magnitude
        elif isinstance(target, (int, float)):
            return self.magnitude >= target
        else:
            return False

    # --------------------------------------------------------------------------
    # Operators
    #

    def __add__(self, other):
        if isinstance(other, Vector):
            return Vector(self.pix + other.pix, "pix", self.win)

    def __sub__(self, other):
        if isinstance(other, Vector):
            return Vector(self.pix - other.pix, "pix", self.win)

    def __mul__(self, other):
        if isinstance(other, Vector):
            return Vector(self.pix * other.pix, "pix", self.win)
        if isinstance(other, (int, float)):
            return Vector(self.pix * other, "pix", self.win)
        if isinstance(other, (list, tuple, np.ndarray)):
            return Vector(self.pix * np.array(other), "pix", self.win)

    def __truediv__(self, other):
        if isinstance(other, Vector):
            return Vector(self.pix / other.pix, "pix", self.win)
        if isinstance(other, (int, float)):
            return Vector(self.pix / other, "pix", self.win)
        if isinstance(other, (list, tuple, np.ndarray)):
            return Vector(self.pix / np.array(other), "pix", self.win)

    # --------------------------------------------------------------------------
    # Class methods and properties
    #

    def copy(self):
        """Create a copy of this object"""
        return self.__copy__()

    def __copy__(self):
        return self.__deepcopy__()

    def __deepcopy__(self):
        return self.__class__(self._requested, self._requestedUnits, self.win)

    @property
    def monitor(self):
        """The monitor used for calculations within this object
        (`~psychopy.monitors.Monitor`).
        """
        return self.win.monitor

    @property
    def dimensions(self):
        """How many dimensions (x, y, z) are specified?"""
        # Run _requested through validator to sanitise it
        value, units = self.validate(self._requested, self._requestedUnits)

        if len(value.shape) == 1:
            # If single value, return number of coords
            return len(value)
        else:
            # If multi value, return number of columns
            return value.shape[1]

    def __len__(self):
        """How many values are specified?"""
        # Run _requested through validator to sanitise it
        value, units = self.validate(self._requested, self._requestedUnits)

        if len(value.shape) == 1:
            # If single value, return 1
            return 1
        else:
            # If multi value, return number of rows
            return value.shape[0]

    @property
    def magnitude(self):
        """Magnitude of vector (i.e. length of the line from vector to (0, 0)
        in pixels).
        """
        return np.hypot3d(*self.pix)

    @property
    def direction(self):
        """Direction of vector (i.e. angle between vector and the horizontal
        plane).
        """
        if self.dimensions < 2:
            # with only 1 dimension, y is essentially zero, so angle is always 0
            return 0.0

        toReturn = []  # store output values

        if self.dimensions >= 2:
            if self.pix[0] != 0.0:  # Angle from x-axis (y is opp, x is adj)
                x = np.degrees(np.arctan(self.pix[1] / self.pix[0]))
            else:
                x = 90.0

            toReturn.append(x)

            if self.pix[1] != 0.0:  # Angle from y-axis (x is opp, y is adj)
                y = np.degrees(np.arctan(self.pix[0] / self.pix[1]))
            else:
                y = 90.0

            toReturn.append(y)

        if self.dimensions == 3:
            # Angle from z-axis (z is opp, hyp(x,y) is adj)
            if np.hypot3d(*self.pix[:2]) != 0.0:
                u = np.hypot3d(*self.pix[:2])
                z = np.degrees(np.arctan(self.pix[2] / u))
            else:
                z = 90.0

            toReturn.append(z)

        return toReturn

    @property
    def pix(self):
        """Values in units of 'pix' (pixels).
        """
        # Check that object is valid
        assert self.valid, (
            u"Could not access pixel value of invalid position/size object")

        # Return cached value if present
        if 'pix' in self._cache:
            return self._cache['pix']
        else:
            raise AttributeError(
                f"Could not retrieve pixel value of Vector object set in "
                f"{self._requestedUnits}")

    @pix.setter
    def pix(self, value):
        # Validate
        value, units = self.validate(value, 'pix')
        # Clear cache and set
        self._cache = {
            'pix': value
        }

    @property
    def deg(self):
        """Values in units of 'deg' (degrees of visual angle).
        """
        # Return cached value if present
        if 'deg' in self._cache:
            return self._cache['deg']
        # Otherwise, do conversion and cache
        self._cache['deg'] = tools.pix2deg(self.pix, self.monitor)
        # Return new cached value
        return self._cache['deg']

    @deg.setter
    def deg(self, value):
        # Validate
        value, units = self.validate(value, 'deg')
        # Convert and set
        self.pix = tools.deg2pix(value, self.monitor)

    @property
    def degFlat(self):
        """Values in units of 'degFlat' (degrees of visual angle corrected for
        screen curvature).

        When dealing with positions/sizes in isolation; 'deg', 'degFlat' and
        'degFlatPos' are synonymous - as the conversion is done at the vertex
        level.
        """
        return self.deg

    @degFlat.setter
    def degFlat(self, value):
        self.deg = value

    @property
    def degFlatPos(self):
        """Values in units of 'degFlatPos'.

        When dealing with positions/sizes in isolation; 'deg', 'degFlat' and
        'degFlatPos' are synonymous - as the conversion is done at the vertex
        level.
        """
        return self.degFlat

    @degFlatPos.setter
    def degFlatPos(self, value):
        self.degFlat = value

    @property
    def cm(self):
        """Values in units of 'cm' (centimeters).
        """
        # Return cached value if present
        if 'cm' in self._cache:
            return self._cache['cm']
        # Otherwise, do conversion and cache
        self._cache['cm'] = tools.pix2cm(self.pix, self.monitor)
        # Return new cached value
        return self._cache['cm']

    @cm.setter
    def cm(self, value):
        # Validate
        value, units = self.validate(value, 'cm')
        # Convert and set
        self.pix = tools.cm2pix(value, self.monitor)

    @property
    def pt(self):
        """Vector coordinates in 'pt' (points).

        Points are commonly used in print media to define text sizes. One point
        is equivalent to 1/72 inches, or around 0.35 mm.
        """
        # Return cached value if present
        if 'pt' in self._cache:
            return self._cache['pt']
        # Otherwise, do conversion and cache
        self._cache['pt'] = self.cm / (2.54 / 72)
        # Return new cached value
        return self._cache['pt']

    @pt.setter
    def pt(self, value):
        # Validate
        value, units = self.validate(value, 'height')
        # Convert and set
        self.cm = value * (2.54 / 72)

    @property
    def norm(self):
        """Value in units of 'norm' (normalized device coordinates).
        """
        # Return cached value if present
        if 'norm' in self._cache:
            return self._cache['norm']
        # Otherwise, do conversion and cache
        buffer = np.ndarray(self.pix.shape, dtype=float)
        for i in range(self.dimensions):
            u = self.win.useRetina + 1
            if len(self) > 1:
                buffer[:, i] = self.pix[:, i] / (self.win.size[i] / u) * 2
            else:
                buffer[i] = self.pix[i] / (self.win.size[i] / u) * 2

        self._cache['norm'] = buffer

        return self._cache['norm']  # return new cached value

    @norm.setter
    def norm(self, value):
        # Validate
        value, units = self.validate(value, 'norm')

        # Convert and set
        buffer = np.ndarray(value.shape, dtype=float)
        for i in range(self.dimensions):
            u = self.win.useRetina + 1
            if len(self) > 1:
                buffer[:, i] = value[:, i] * (self.win.size[i] / u) / 2
            else:
                buffer[i] = value[i] * (self.win.size[i] / u) / 2

        self.pix = buffer

    @property
    def height(self):
        """Value in units of 'height' (normalized to the height of the window).
        """
        # Return cached value if present
        if 'height' in self._cache:
            return self._cache['height']
        # Otherwise, do conversion and cache
        self._cache['height'] = \
            self.pix / (self.win.size[1] / (self.win.useRetina + 1))
        # Return new cached value
        return self._cache['height']

    @height.setter
    def height(self, value):
        # Validate
        value, units = self.validate(value, 'height')
        # Convert and set
        self.pix = value * (self.win.size[1] / (self.win.useRetina + 1))


class Position(Vector):
    """Class representing a position vector.

    This class is used to specify the location of a point within some
    coordinate system (e.g., `(x, y)`).

    Parameters
    ----------
    value : ArrayLike
        Array of coordinates representing positions within a coordinate system.
        Positions are specified in a similar manner to `~psychopy.layout.Vector`
        as either 1xN for single vectors, and Nx2 or Nx3 for multiple positions.
    units : str or None
        Units which `value` has been specified in. Applicable values are
        `'pix'`, `'deg'`, `'degFlat'`, `'degFlatPos'`, `'cm'`, `'pt'`, `'norm'`,
        `'height'`, or `None`.
    win : `~psychopy.visual.Window` or None
        Window associated with this position. This value must be specified if
        you wish to map positions to coordinate systems that require additional
        information about the monitor the window is being displayed on.

    """
    def __init__(self, value, units, win=None):
        Vector.__init__(self, value, units, win)


class Size(Vector):
    """Class representing a size.

    Parameters
    ----------
    value : ArrayLike
        Array of values representing size axis-aligned bounding box within a
        coordinate system. Sizes are specified in a similar manner to
        `~psychopy.layout.Vector` as either 1xN for single vectors, and Nx2 or
        Nx3 for multiple positions.
    units : str or None
        Units which `value` has been specified in. Applicable values are
        `'pix'`, `'deg'`, `'degFlat'`, `'degFlatPos'`, `'cm'`, `'pt'`, `'norm'`,
        `'height'`, or `None`.
    win : `~psychopy.visual.Window` or None
        Window associated with this size object. This value must be specified if
        you wish to map sizes to coordinate systems that require additional
        information about the monitor the window is being displayed on.

    """
    def __init__(self, value, units, win=None):
        Vector.__init__(self, value, units, win)


class Vertices:
    """Class representing an array of vertices.

    Parameters
    ----------
    verts : ArrayLike
        Array of coordinates specifying the locations of vertices.
    obj : object or None
    size : ArrayLike or None
        Scaling factors for vertices along each dimension.
    pos : ArrayLike or None
        Offset for vertices along each dimension.
    units : str or None
        Units which `verts` has been specified in. Applicable values are
        `'pix'`, `'deg'`, `'degFlat'`, `'degFlatPos'`, `'cm'`, `'pt'`, `'norm'`,
        `'height'`, or `None`.
    flip : ArrayLike or None
        Array of boolean values specifying which dimensions to flip/mirror.
        Mirroring is applied prior to any other transformation.
    anchor : str or None
        Anchor location for vertices, specifies the origin for the vertices.

    """
    def __init__(self, verts, obj=None, size=None, pos=None, units=None,
                 flip=None, anchor=None):

        if obj is None and pos is None and size is None:
            raise ValueError(
                "Vertices array needs either an object or values for pos and "
                "size.")

        # Store object
        self.obj = obj

        # Store size and pos
        self._size = size
        self._pos = pos
        self._units = units
        self.flip = flip  # store flip
        self.anchor = anchor  # set anchor

        # Convert to numpy array
        verts = np.array(verts)

        # Make sure it's coercible to a Nx2 or nxNx2 numpy array
        assert (3 >= len(verts.shape) >= 2) and (verts.shape[-1] == 2), (
            f"Expected vertices to be coercible to a Nx2 or nxNx2 numpy array, not {verts.shape}"
        )

        # Store base vertices
        self.base = verts

    def __repr__(self):
        """If vertices object is printed, it will display its class and value.
        """
        if self:
            return (
                f"<psychopy.layout.{self.__class__.__name__}: "
                f"{np.round(self.base, 3)} * "
                f"{np.round(self.obj._size.pix, 3)} + "
                f"{np.round(self.obj._pos.pix, 3)}>")
        else:
            return "<psychopy.layout.{self.__class__.__name__}: Invalid>"

    @property
    def pos(self):
        """Positional offset of the vertices (`~psychopy.layout.Vector` or
        ArrayLike)."""
        if isinstance(self._pos, Vector):
            return self._pos
        if hasattr(self.obj, "_pos"):
            return self.obj._pos
        else:
            raise AttributeError(
                f"Could not derive position from object {self.obj} as object "
                f"does not have a position attribute.")

    @property
    def size(self):
        """Scaling factors for vertices (`~psychopy.layout.Vector` or
        ArrayLike)."""
        if isinstance(self._size, Vector):
            return self._size
        if hasattr(self.obj, "_size"):
            return self.obj._size
        else:
            raise AttributeError(
                f"Could not derive size from object {self.obj} as object does "
                f"not have a size attribute.")

    @property
    def units(self):
        """Units which the vertices are specified in (`str`).
        """
        if hasattr(self, "_units") and self._units is not None:
            return self._units
        if hasattr(self, "obj") and hasattr(self.obj, "units"):
            return self.obj.units

    @property
    def flip(self):
        """1x2 array for flipping vertices along each axis; set as `True` to
        flip or `False` to not flip (`ArrayLike`).

        If set as a single value, will duplicate across both axes. Accessing the
        protected attribute (`._flip`) will give an array of 1s and -1s with
        which to multiply vertices.
        """
        # Get base value
        if hasattr(self, "_flip"):
            flip = self._flip
        else:
            flip = np.array([[False, False]])
        # Convert from boolean
        return flip == -1

    @flip.setter
    def flip(self, value):
        if value is None:
            value = False

        # Convert to 1x2 numpy array
        value = np.array(value)
        value = np.resize(value, (1, 2))

        # Ensure values were bool
        assert value.dtype == bool, (
            "Flip values must be either a boolean (True/False) or an array of "
            "booleans")

        # Set as multipliers rather than bool
        self._flip = np.array([[
            -1 if value[0, 0] else 1,
            -1 if value[0, 1] else 1,
        ]])
        self._flipHoriz, self._flipVert = self._flip[0]

    @property
    def flipHoriz(self):
        """Apply horizontal mirroring (`bool`)?
        """
        return self.flip[0][0]

    @flipHoriz.setter
    def flipHoriz(self, value):
        self.flip = [value, self.flip[0, 1]]

    @property
    def flipVert(self):
        """Apply vertical mirroring (`bool`)?
        """
        return self.flip[0][1]

    @flipVert.setter
    def flipVert(self, value):
        self.flip = [self.flip[0, 0], value]

    @property
    def anchor(self):
        """Anchor location (`str`).

        Possible values are on of `'top'`, `'bottom'`, `'left'`, `'right'`,
        `'center'`. Combinations of these values may also be specified (e.g.,
        `'top_center'`, `'center-right'`, `'topleft'`, etc. are all valid).
        """
        if hasattr(self, "_anchorX") and hasattr(self, "_anchorY"):
            # If set, return set values
            return self._anchorX, self._anchorY

        # Otherwise, assume center
        return "center", "center"

    @anchor.setter
    def anchor(self, anchor):
        # Set defaults
        self._anchorY = None
        self._anchorX = None
        # Look for unambiguous terms first (top, bottom, left, right)
        if 'top' in str(anchor):
            self._anchorY = 'top'
        elif 'bottom' in str(anchor):
            self._anchorY = 'bottom'
        if 'right' in str(anchor):
            self._anchorX = 'right'
        elif 'left' in str(anchor):
            self._anchorX = 'left'
        # Then 'center' can apply to either axis that isn't already set
        if self._anchorX is None:
            self._anchorX = 'center'
        if self._anchorY is None:
            self._anchorY = 'center'

    @property
    def anchorAdjust(self):
        """Map anchor values to numeric vertices adjustments.
        """
        return [_anchorAliases[a] for a in self.anchor]

    def getas(self, units):
        assert units in unitTypes, f"Unrecognised unit type '{units}'"
        # Start with base values
        verts = self.base.copy()
        verts = verts.astype(float)
        # Apply anchor
        verts += self.anchorAdjust
        # Apply size
        if self.size is None:
            raise ValueError(
                u"Cannot not calculate absolute positions of vertices without "
                u"a size attribute")
        verts *= getattr(self.size, units)
        # Apply flip
        verts *= self._flip
        # Apply pos
        if self.pos is None:
            raise ValueError(
                u"Cannot not calculate absolute positions of vertices without "
                u"a pos attribute")
        verts += getattr(self.pos, units)

        return verts

    def setas(self, value, units):
        assert units in unitTypes, f"Unrecognised unit type '{units}'"
        # Enforce numpy
        value = np.array(value, dtype=float)
        # Account for size
        if self.size is None:
            raise ValueError(
                u"Cannot not calculate absolute positions of vertices without "
                u"a size attribute")
        value /= getattr(self.size, units)
        # Account for flip
        value *= self._flip
        # Account for pos
        if self.pos is None:
            raise ValueError(
                u"Cannot not calculate absolute positions of vertices without "
                u"a pos attribute")

        value -= getattr(self.pos, units)
        self.base = value  # apply

    @property
    def pix(self):
        """Get absolute positions of vertices in 'pix' units.
        """
        # If correcting for screen curve, use the old functions
        if self.units == 'degFlat':
            return tools._degFlat2pix(
                self.base * self.obj.size, self.obj.pos, self.obj.win)
        elif self.units == 'degFlatPos':
            return tools._degFlatPos2pix(
                self.base * self.obj.size, self.obj.pos, self.obj.win)
        else:
            # Otherwise, use standardised method
            return self.getas('pix')

    @pix.setter
    def pix(self, value):
        self.setas(value, 'pix')

    @property
    def deg(self):
        """Get absolute positions of vertices in 'deg' units.
        """
        return self.getas('deg')

    @deg.setter
    def deg(self, value):
        self.setas(value, 'deg')

    @property
    def degFlat(self):
        """Get absolute positions of vertices in 'degFlat' units.
        """
        return self.getas('degFlat')

    @degFlat.setter
    def degFlat(self, value):
        cm = tools.deg2cm(value, self.obj.win.monitor, correctFlat=True)
        self.setas(cm, 'cm')

    @property
    def cm(self):
        """Get absolute positions of vertices in 'cm' units.
        """
        return self.getas('cm')

    @cm.setter
    def cm(self, value):
        self.setas(value, 'cm')

    @property
    def norm(self):
        """Get absolute positions of vertices in 'norm' units.
        """
        return self.getas('norm')

    @norm.setter
    def norm(self, value):
        self.setas(value, 'norm')

    @property
    def height(self):
        """Get absolute positions of vertices in 'height' units.
        """
        return self.getas('height')

    @height.setter
    def height(self, value):
        self.setas(value, 'height')


if __name__ == "__main__":
    pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

# DONE: add+commit before push
# DONE:  add .gitignore file. Added when opening a repo without one
# DONE: fork+sync doesn't yet fork the project first
# DONE: rather than clone into a folder with files we should init/add/push
#
# TODO: after clone, remember this folder for next file-open call
# TODO: user dlg could/should be local not a browser
# TODO: syncProject() doesn't handle case of a local git pushing to new gitlab
# TODO: if more than one remote then offer options

from psychopy.projects import pavlovia
from .functions import *
from .project import ProjectEditor, syncProject
from ._base import PavloviaMiniBrowser
from . import menu, project, search, toolbar



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Load and play sounds

We have used a number of different Python libraries ("backends") for generating
sounds in PsychoPy. We started with `Pygame`, then tried `pyo` and `sounddevice`
but we now strongly recommend you use the PTB setting. That uses the
`PsychPortAudio`_ engine, written by Mario Kleiner for `Psychophysics Toolbox`_.

With the PTB backend you get some options about how aggressively you want to try
for low latency, and there is also an option to pre-schedule a sound to play at
a given time in the future.

By default PsychoPy will try to use the following Libs, in this order, for
sound reproduction but you can alter the order in
preferences > hardware > audioLib:
    ['sounddevice', 'pygame', 'pyo']
For portaudio-based backends (all except for pygame) there is also a
choice of the underlying sound driver (e.g. ASIO, CoreAudio etc).

After importing sound, the sound lib and driver being used will be stored as::
    `psychopy.sound.audioLib`
    `psychopy.sound.audioDriver`

.. PTB

.. _PsychPortAudio: http://psychtoolbox.org/docs/PsychPortAudio-Open
.. _Psychophysics Toolbox: http://psychtoolbox.org
"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

__all__ = []

import sys
import os
import traceback
from psychopy import logging, prefs, constants
from .exceptions import DependencyError, SoundFormatError
from .audiodevice import *
from .audioclip import *  # import objects related to AudioClip

# import microphone if possible
try:
    from .microphone import *  # import objects related to the microphone class
except ImportError as err:
    formatted_tb = ''.join(traceback.format_exception(type(err), err, err.__traceback__))
    logging.error("Failed to import psychopy.sound.microphone. Mic recordings will not be"
                  "possible on this machine. For details see stack trace below:\n"
                  f"{formatted_tb}")

# import transcription if possible
try:
    from .transcribe import *  # import transcription engine stuff
except ImportError as err:
    formatted_tb = ''.join(traceback.format_exception(type(err), err, err.__traceback__))
    logging.error("Failed to import psychopy.sound.transcribe. Transcription will not be"
                  "possible on this machine. For details see stack trace below:\n"
                  f"{formatted_tb}")

pyoSndServer = None
Sound = None
audioLib = None
audioDriver = None
bits32 = sys.maxsize == 2 ** 32

_audioLibs = ['PTB', 'sounddevice', 'pyo', 'pysoundcard', 'pygame']
failed = []

# check if this is being imported on Travis (has no audio card)
travisCI = bool(str(os.environ.get('TRAVIS')).lower() == 'true')
if travisCI:
    # for sounddevice we built in some TravisCI protection but not in pyo
    prefs.hardware['audioLib'] = ['sounddevice']

if isinstance(prefs.hardware['audioLib'], str):
    prefs.hardware['audioLib'] = [prefs.hardware['audioLib']]
for thisLibName in prefs.hardware['audioLib']:
    try:
        if thisLibName.lower() == 'ptb':
            from . import backend_ptb as backend
            Sound = backend.SoundPTB
            audioDriver = backend.audioDriver
        elif thisLibName == 'pyo':
            from . import backend_pyo as backend
            Sound = backend.SoundPyo
            pyoSndServer = backend.pyoSndServer
            audioDriver = backend.audioDriver
        elif thisLibName == 'sounddevice':
            from . import backend_sounddevice as backend
            Sound = backend.SoundDeviceSound
        elif thisLibName == 'pygame':
            from . import backend_pygame as backend
            Sound = backend.SoundPygame
        elif thisLibName == 'pysoundcard':
            from . import backend_pysound as backend
            Sound = backend.SoundPySoundCard
        else:
            msg = ("audioLib pref should be one of {!r}, not {!r}"
                   .format(_audioLibs, thisLibName))
            raise ValueError(msg)
        # if we got this far we were successful in loading the lib
        audioLib = thisLibName
        init = backend.init
        if hasattr(backend, 'getDevices'):
            getDevices = backend.getDevices
        logging.info('sound is using audioLib: %s' % audioLib)
        break
    except DependencyError as e:
        failed.append(thisLibName.lower())
        msg = '%s audio lib was requested but not loaded: %s'
        logging.warning(msg % (thisLibName, sys.exc_info()[1]))
        continue  # to try next audio lib

if audioLib is None:
    raise DependencyError(
            "No sound libs could be loaded. Tried: {}\n"
            "Check whether the necessary sound libs are installed"
            .format(prefs.hardware['audioLib']))
elif audioLib.lower() != 'ptb':
    if not bits32 and 'ptb' not in failed:
        # Could be running PTB, just aren't?
        logging.warning("We strongly recommend you activate the PTB sound "
                        "engine in PsychoPy prefs as the preferred audio "
                        "engine. Its timing is vastly superior. Your prefs "
                        "are currently set to use {} (in that order)."
                        .format(prefs.hardware['audioLib']))
    else:  # Can't run PTB anyway due to Py2 or 32bit system
        logging.warning("For experiments that use audio stimuli, timing will "
                        "be much better if you upgrade your PsychoPy "
                        "installation to a 64bit Python3 installation and use "
                        "the PTB backend.")


# function to set the device (if current lib allows it)
def setDevice(dev, kind=None):
    """Sets the device to be used for new streams being created.

    :param dev: the device to be used (name, index or sounddevice.device)
    :param kind: one of [None, 'output', 'input']
    """
    if not hasattr(backend, 'defaultOutput'):
        raise IOError("Attempting to SetDevice (audio) but not supported by "
                      "the current audio library ({!r})".format(audioLib))
    if hasattr(dev,'name'):
        dev = dev['name']
    if kind is None:
        backend.defaultInput = backend.defaultOutput = dev
    elif kind == 'input':
        backend.defaultInput = dev
    elif kind == 'output':
        backend.defaultOutput = dev
    else:
        if travisCI:  # travisCI doesn't have any audio devices at all. Ignore
            return
        else:
            raise TypeError("`kind` should be one of [None, 'output', 'input']"
                            "not {!r}".format(kind))

# Set the device according to user prefs (if current lib allows it)
deviceNames = []
if hasattr(backend, 'defaultOutput'):
    pref = prefs.hardware['audioDevice']
    # is it a list or a simple string?
    if type(prefs.hardware['audioDevice'])==list:
        # multiple options so use zeroth
        dev = prefs.hardware['audioDevice'][0]
    else:
        # a single option
        dev = prefs.hardware['audioDevice']
    # is it simply "default" (do nothing)
    if dev=='default' or travisCI:
        pass  # do nothing
    elif dev not in backend.getDevices(kind='output'):
        deviceNames = sorted(backend.getDevices(kind='output').keys())
        logging.warn(u"Requested audio device '{}' that is not available on "
                        "this hardware. The 'audioDevice' preference should be one of "
                        "{}".format(dev, deviceNames))
    else:
        setDevice(dev, kind='output')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
from psychopy import event, core
from psychopy.preferences import prefs
from psychopy.visual import Window

from pyglet.window.key import (MOD_SHIFT,
                               MOD_CTRL,
                               MOD_ALT,
                               MOD_CAPSLOCK,
                               MOD_NUMLOCK,
                               MOD_WINDOWS,
                               MOD_COMMAND,
                               MOD_OPTION,
                               MOD_SCROLLLOCK)


@pytest.mark.keyboard
class TestKeyboardEvents():

    def test_keyname(self):
        """Test that a key name is correctly returned."""
        event._onPygletKey('a', 0, emulated=True)
        keys = event.getKeys()
        assert len(keys) == 1
        assert keys[0] == 'a'

    def test_modifiers(self):
        """Test that key modifier flags are correctly returned"""
        event._onPygletKey('a', MOD_CTRL|MOD_SHIFT, emulated=True)
        keys = event.getKeys(modifiers=True)
        assert len(keys) == 1
        assert len(keys[0]) == 2
        assert keys[0][0] == 'a'
        assert keys[0][1]['ctrl']
        assert keys[0][1]['shift']
        assert not keys[0][1]['alt']

    def test_timestamp(self):
        """Test that a keypress timestamp is correctly returned"""
        event._onPygletKey('a', 0, emulated=True)
        keys = event.getKeys(timeStamped=True)
        assert len(keys) == 1
        assert len(keys[0]) == 2
        assert keys[0][0] == 'a'
        assert isinstance(keys[0][1], float)
        assert keys[0][1] > 0.0

    def test_modifiers_and_timestamp(self):
        """Test that both key modifiers and timestamp are returned"""
        event._onPygletKey('a', MOD_ALT, emulated=True)
        keys = event.getKeys(modifiers=True, timeStamped=True)
        assert len(keys) == 1
        assert len(keys[0]) == 3
        assert keys[0][0] == 'a'
        assert keys[0][1]['alt']
        assert isinstance(keys[0][2], float)

    def test_invalid_modifiers(self):
        """Modifiers must be integers."""
        key = 'a'
        modifiers = None

        with pytest.raises(ValueError):
            event._onPygletKey(key, modifiers, emulated=True)

    def test_german_characters(self):
        """Test that event can handle German characters"""
        # 824633720832 = ö as pyglet symbol string
        # need to use emulated = False to execute the lines that actually
        # fix the German characters handling
        event._onPygletKey(824633720832, 0, emulated=False)
        event._onPygletText('ö', emulated=True)
        keys = event.getKeys(modifiers=False, timeStamped=True)
        assert len(keys) == 1
        assert len(keys[0]) == 2
        assert keys[0][0] == 'ö'
        assert isinstance(keys[0][1], float)

    def test_german_characters_with_modifiers(self):
        """Test that event can handle German characters with modifiers"""
        # 824633720832 = ö as pyglet symbol string
        # need to use emulated = False to execute the lines that actually
        # fix the German characters handling
        event._onPygletKey(824633720832, MOD_SHIFT | MOD_SCROLLLOCK, emulated=False)
        event._onPygletText('ö', emulated=True)
        keys = event.getKeys(modifiers=True, timeStamped=True)
        assert len(keys) == 1
        assert len(keys[0]) == 3
        assert keys[0][0] == 'ö'
        assert keys[0][1]['shift']
        assert keys[0][1]['scrolllock']
        assert isinstance(keys[0][2], float)


@pytest.mark.keyboard
class TestGLobalEventKeys():
    @classmethod
    def setup_class(self):
        self.win = Window([128, 128], winType='pyglet', pos=[50, 50], autoLog=False)

    @classmethod
    def teardown_class(self):
        self.win.close()

    def setup_method(self, test_method):
        # Disable auto-creation of shutdown key.
        prefs.general['shutdownKey'] = ''

    def _func(self, *args, **kwargs):
        return [args, kwargs]

    def test_shutdownKey_prefs(self):
        key = 'escape'
        modifiers = ('ctrl', 'alt')

        prefs.general['shutdownKey'] = key
        prefs.general['shutdownKeyModifiers'] = modifiers

        global_keys = event._GlobalEventKeys()
        e = list(global_keys)[0]

        assert key, modifiers == e
        assert global_keys[e].func == core.quit

    def test_add(self):
        key = 'a'
        func = self._func

        global_keys = event._GlobalEventKeys()
        global_keys.add(key=key, func=func)

        assert global_keys[key, ()].func == func
        assert global_keys[key, ()].name == func.__name__

    def test_add_key_twice(self):
        key = 'a'
        func = self._func
        global_keys = event._GlobalEventKeys()
        global_keys.add(key=key, func=func)

        with pytest.raises(ValueError):
            global_keys.add(key=key, func=func)

    def test_add_name(self):
        key = 'a'
        name = 'foo'
        func = self._func

        global_keys = event._GlobalEventKeys()
        global_keys.add(key=key, func=func, name=name)
        assert global_keys[key, ()].name == name

    def test_add_args(self):
        key = 'a'
        func = self._func
        args = (1, 2, 3)

        global_keys = event._GlobalEventKeys()
        global_keys.add(key=key, func=func, func_args=args)

        assert global_keys[key, ()].func_args == args

    def test_add_kwargs(self):
        key = 'a'
        func = self._func
        kwargs = dict(foo=1, bar=2)

        global_keys = event._GlobalEventKeys()
        global_keys.add(key=key, func=func, func_kwargs=kwargs)

        assert global_keys[key, ()].func_kwargs == kwargs

    def test_add_args_and_kwargs(self):
        key = 'a'
        func = self._func
        args = (1, 2, 3)
        kwargs = dict(foo=1, bar=2)

        global_keys = event._GlobalEventKeys()
        global_keys.add(key=key, func=func, func_args=args,
                        func_kwargs=kwargs)

        assert global_keys[key, ()].func_args == args
        assert global_keys[key, ()].func_kwargs == kwargs

    def test_add_invalid_key(self):
        key = 'foo'
        func = self._func
        global_keys = event._GlobalEventKeys()

        with pytest.raises(ValueError):
            global_keys.add(key=key, func=func)

    def test_add_invalid_modifiers(self):
        key = 'a'
        modifiers = ('foo', 'bar')
        func = self._func
        global_keys = event._GlobalEventKeys()

        with pytest.raises(ValueError):
            global_keys.add(key=key, modifiers=modifiers, func=func)

    def test_remove(self):
        keys = ['a', 'b', 'c']
        modifiers = ('ctrl',)
        func = self._func

        global_keys = event._GlobalEventKeys()
        [global_keys.add(key=key, modifiers=modifiers, func=func)
         for key in keys]

        global_keys.remove(keys[0], modifiers)
        with pytest.raises(KeyError):
            _ = global_keys[keys[0], modifiers]

    def test_remove_modifiers_list(self):
        key = 'a'
        modifiers = ['ctrl', 'alt']
        func = self._func

        global_keys = event._GlobalEventKeys()
        global_keys.add(key=key, modifiers=modifiers, func=func)

        global_keys.remove(key, modifiers)
        with pytest.raises(KeyError):
            _ = global_keys[key, modifiers]

    def test_remove_invalid_key(self):
        key = 'a'
        global_keys = event._GlobalEventKeys()

        with pytest.raises(KeyError):
            global_keys.remove(key)

    def test_remove_all(self):
        keys = ['a', 'b', 'c']
        func = self._func

        global_keys = event._GlobalEventKeys()
        [global_keys.add(key=key, func=func) for key in keys]

        global_keys.remove('all')
        assert len(global_keys) == 0

    def test_getitem(self):
        key = 'escape'
        modifiers = ('ctrl', 'alt')
        func = self._func

        global_keys = event._GlobalEventKeys()
        global_keys.add(key=key, modifiers=modifiers, func=func)

        assert global_keys[key, modifiers] == global_keys._events[key, modifiers]

    def test_getitem_string(self):
        key = 'escape'
        func = self._func

        global_keys = event._GlobalEventKeys()
        global_keys.add(key=key, func=func)

        assert global_keys[key] == global_keys._events[key, ()]

    def test_getitem_modifiers_list(self):
        key = 'escape'
        modifiers = ['ctrl', 'alt']
        func = self._func

        global_keys = event._GlobalEventKeys()
        global_keys.add(key=key, modifiers=modifiers, func=func)

        assert (global_keys[key, modifiers] ==
                global_keys._events[key, tuple(modifiers)])

    def test_setitem(self):
        keys = 'a'
        modifiers = ()
        global_keys = event._GlobalEventKeys()

        with pytest.raises(NotImplementedError):
            global_keys[keys, modifiers] = None

    def test_delitem(self):
        key = 'escape'
        modifiers = ('ctrl', 'alt')
        func = self._func

        global_keys = event._GlobalEventKeys()
        global_keys.add(key=key, modifiers=modifiers, func=func)

        del global_keys[key, modifiers]
        with pytest.raises(KeyError):
            _ = global_keys[key, modifiers]

    def test_delitem_string(self):
        key = 'escape'
        func = self._func

        global_keys = event._GlobalEventKeys()
        global_keys.add(key=key, func=func)

        del global_keys[key]
        with pytest.raises(KeyError):
            _ = global_keys[key]

    def test_len(self):
        prefs.general['shutdownKey'] = ''
        key = 'escape'
        func = self._func

        global_keys = event._GlobalEventKeys()
        assert len(global_keys) == 0

        global_keys.add(key=key, func=func)
        assert len(global_keys) == 1

        del global_keys[key, ()]
        assert len(global_keys) == 0

    def test_event_processing(self):
        key = 'a'
        modifiers = 0
        func = self._func
        args = (1, 2, 3)
        kwargs = dict(foo=1, bar=2)

        event.globalKeys.add(key=key, func=func, func_args=args,
                             func_kwargs=kwargs)

        r = event._process_global_event_key(key, modifiers)
        assert r[0] == args
        assert r[1] == kwargs

    def test_index_keys(self):
        key = 'escape'
        modifiers = ('ctrl', 'alt')
        func = self._func

        global_keys = event._GlobalEventKeys()
        global_keys.add(key=key, modifiers=modifiers, func=func)

        index_key = list(global_keys.keys())[-1]
        assert index_key.key == key
        assert index_key.modifiers == modifiers

    def test_numlock(self):
        key = 'a'
        modifiers = ('numlock',)
        func = self._func

        global_keys = event._GlobalEventKeys()

        with pytest.raises(ValueError):
            global_keys.add(key=key, modifiers=modifiers, func=func)


if __name__ == '__main__':
    import pytest
    pytest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo of TrialHandler

The contents of this file are in the public domain.

"""

from random import random

from psychopy import data

# create your list of stimuli
# NB as of version 1.62 you could simply import an excel spreadsheet with this
# using data.importConditions('someFile.xlsx')
stimList = []
for ori in range(90, 180, 30):
    for sf in [0.5, 1.0, 2.0]:
        # append a python 'dictionary' to the list
        stimList.append({'sf': sf, 'ori': ori})

# organize them with the trial handler
trials = data.TrialHandler(stimList, 10,
                           extraInfo={'participant': "Nobody", 'session':'001'})

# run the experiment
nDone = 0
for thisTrial in trials:  # handler can act like a for loop
    # simulate some data
    thisReactionTime = random() + float(thisTrial['sf']) / 2.0
    thisChoice = round(random())
    trials.data.add('RT', thisReactionTime)  # add the data to our set
    trials.data.add('choice', thisChoice)
    nDone += 1  # just for a quick reference

    msg = 'trial %i had position %s in the list (sf=%.1f)'
    print(msg % (nDone, trials.thisIndex, thisTrial['sf']))

# After the experiment, print a new line
print('\n')

# Write summary data to screen
trials.printAsText(stimOut=['sf', 'ori'],
                   dataOut=['RT_mean', 'RT_std', 'choice_raw'])

# Write summary data to a text file ...
trials.saveAsText(fileName='testData',
                  stimOut=['sf', 'ori'],
                  dataOut=['RT_mean', 'RT_std', 'choice_raw'])

# ... or an xlsx file (which supports sheets)
trials.saveAsExcel(fileName='testData',
                   sheetName='rawData',
                   stimOut=['sf', 'ori'],
                   dataOut=['RT_mean', 'RT_std', 'choice_raw'])

# Save a copy of the whole TrialHandler object, which can be reloaded later to
# re-create the experiment.
trials.saveAsPickle(fileName='testData')

# Wide format is useful for analysis with R or SPSS.
df = trials.saveAsWideText('testDataWide.txt')



# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).

"""
ioHub Eye Tracker Online Sample Event Parser

WORK IN PROGRESS - VERY EXPERIMENTAL

Copyright (C) 2012-2014 iSolver Software Solutions
Distributed under the terms of the GNU General Public License
(GPL version 3 or any later version).

.. moduleauthor:: Sol Simpson <sol@isolver-software.com>
.. fileauthor:: Sol Simpson <sol@isolver-software.com>

NOTES:

* The parser is designed to work with monocular and binocular eye data,
but only binocular input samples have been tested so far.
* If binocular input samples are being used, they are converted to monocular
samples for parsing. If both left and right eye position data is available
for a sample, the positions are averaged together. If only one of the two eyes
has valid data, then that eye data is used for the sample. So the only case
where a sample will be tagged as missing data is when both eyes do not have
valid eye position / pupil size data.

POSITION_FILTER and VELOCITY_FILTER can be set to one of the following event
field filter types. Example values for any input arguments are given. The filter
is selected by giving the filter class name followed by a dictionary of values
to use for the filter. Valid filter options depend on the filter selected.

eventfilters.MovingWindowFilter
--------------------------------

This is a standard averaging filter. Any samples within the window buffer are
simply averaged together to give the filtered value for a given sample.

Parameters:
    * length: The size of the moving window in samples. Minimum of 2 required.
    * knot_pos: The index within the moving window that should be used to extract
                a sample from and apply the current window filtered value.

Example:

POSITION_FILTER = eventfilters.MovingWindowFilter, {length: 3, knot_pos:'center'}

Applies the MovingWindowFilter to x and y gaze data fields of eye samples. The
window size is three, and each sample position is filtered using data from the
previous and next samples as well as itself.

eventfilters.PassThroughFilter
---------------------------------

A NULL filter. In other words, the filter does not do any filtering.

Parameters: None

Example:

VELOCITY_FILTER = eventfilters.PassThroughFilter, {}

Velocity data is calculated from (filtered) sample positions, but is not
filtered itself.

eventfilters.MedianFilter
-----------------------------

MedianFilter applies the median value of the filter window to the knot_pos
window sample.

Parameters:
    * length: The size of the moving window in samples. Minimum of 3 is
      required and the length must be odd.
    * knot_pos: The index within the moving window that should be used to extract
                a sample from and apply the current window filtered value.
Example:

POSITION_FILTER = eventfilters.MedianFilter, {length: 3, knot_pos: 0}

Sample position fields are filtered by the median value of three samples, those
being the current sample and the two following samples (so the current sample is
at index 0.

eventfilters.WeightedAverageFilter
-----------------------------------

WeightedAverageFilter is similar to the standard MovingWindowFilter field filter,
however each element in the window is assigned a weighting factor that is used
during averaging.

Parameters:
    * weights: A list of weights to be applied to the window values. The window
    length is == len(weights). The weight values are all normalized to sum to 1
    before use in the filter. For example, a weight list of (25,50,25) will be
    converted to (0.25,0.50,0.25) for use in the filter, with window value index
    i being multiplied by weith list index i.
    * knot_pos: The index within the moving window that should be used to extract
                a sample from and apply the current window filtered value.

Example:

VELOCITY_FILTER = eventfilters.WeightedAverageFilter, {weights: (25,50,25), knot_pos: 1}

A weighted average window filter will be applied to x and y velocity fields.
The length of the window is 3 samples, and the filtered sample index retrieved
is 1, the same as using 'center' in this case. The filtered sample index will
count toward 1/2 the weighted average, with the previous and next samples
contributing 1/4 of the weighted average each.

eventfilters.StampFilter
--------------------------

A variant of the filter proposed by Dr. David Stampe (1993 ???). A window of
length 3 is used, with the knot_pos centered, or at index 1. If the current
3 values in the window list are monotonic, then the sample is not filtered.
If the values are non-monotonic, then v[1] = (v[0]+v[2])/2.0

Parameters:
    * levels: The number of iterations (recursive) that should be applied to the
              windowed data. Minimum value is 1. The number of levels equals
              the number of samples the filtered sample will be delayed
              compared to the non filtered sample time.

Example:

POSITION_FILTER = eventfilters.StampFilter, {level: 1}

Data is filtered once, similar to what a 'normal' filter level would be in the
  eyelink<tm> system. Level = 2 would be similar to the 'extra' filter level
  setting of eyelink<tm>.
"""
import numpy as np
from ....constants import EventConstants
from ....errors import print2err
from ... import DeviceEvent, eventfilters
from collections import OrderedDict
from ....util.visualangle import VisualAngleCalc

MONOCULAR_EYE_SAMPLE = EventConstants.MONOCULAR_EYE_SAMPLE
BINOCULAR_EYE_SAMPLE = EventConstants.BINOCULAR_EYE_SAMPLE
FIXATION_START = EventConstants.FIXATION_START
FIXATION_END = EventConstants.FIXATION_END
SACCADE_START = EventConstants.SACCADE_START
SACCADE_END = EventConstants.SACCADE_END
BLINK_START = EventConstants.BLINK_START
BLINK_END = EventConstants.BLINK_END

NO_EYE = 0
LEFT_EYE = 1
RIGHT_EYE = 2
BOTH_EYE = 3


class EyeTrackerEventParser(eventfilters.DeviceEventFilter):

    def __init__(self, **kwargs):
        eventfilters.DeviceEventFilter.__init__(self, **kwargs)
        self.sample_type = None
        self.io_sample_class = None
        self.io_event_ix = None
        self.last_valid_sample = None
        self.last_sample = None
        self.invalid_samples_run = []
        self._last_parser_sample = None
        self.open_parser_events = OrderedDict()
        self.convertEvent = None
        self.isValidSample = None
        self.vel_thresh_history_dur = kwargs.get(
            'adaptive_vel_thresh_history', 3.0)
        position_filter = kwargs.get('position_filter')
        velocity_filter = kwargs.get('velocity_filter')
        display_device = kwargs.get('display_device')
        sampling_rate = kwargs.get('sampling_rate')

        if position_filter:
            pos_filter_class_name = position_filter.get(
                'name', 'PassThroughFilter')
            pos_filter_class = getattr(eventfilters, pos_filter_class_name)
            del position_filter['name']
            pos_filter_kwargs = position_filter
        else:
            pos_filter_class, pos_filter_kwargs = eventfilters.PassThroughFilter, {}

        if velocity_filter:
            vel_filter_class_name = position_filter.get(
                'name', 'PassThroughFilter')
            vel_filter_class = getattr(eventfilters, vel_filter_class_name)
            del velocity_filter['name']
            vel_filter_kwargs = velocity_filter
        else:
            vel_filter_class, vel_filter_kwargs = eventfilters.PassThroughFilter, {}

        self.adaptive_x_vthresh_buffer = np.zeros(
            self.vel_thresh_history_dur * sampling_rate)
        self.x_vthresh_buffer_index = 0
        self.adaptive_y_vthresh_buffer = np.zeros(
            self.vel_thresh_history_dur * sampling_rate)
        self.y_vthresh_buffer_index = 0

        pos_filter_kwargs['event_type'] = MONOCULAR_EYE_SAMPLE
        pos_filter_kwargs['inplace'] = True
        pos_filter_kwargs['event_field_name'] = 'angle_x'
        self.x_position_filter = pos_filter_class(**pos_filter_kwargs)
        pos_filter_kwargs['event_field_name'] = 'angle_y'
        self.y_position_filter = pos_filter_class(**pos_filter_kwargs)

        vel_filter_kwargs['event_type'] = MONOCULAR_EYE_SAMPLE
        vel_filter_kwargs['inplace'] = True
        vel_filter_kwargs['event_field_name'] = 'velocity_x'
        self.x_velocity_filter = vel_filter_class(**vel_filter_kwargs)
        vel_filter_kwargs['event_field_name'] = 'velocity_y'
        self.y_velocity_filter = vel_filter_class(**vel_filter_kwargs)
        vel_filter_kwargs['event_field_name'] = 'velocity_xy'
        self.xy_velocity_filter = vel_filter_class(**vel_filter_kwargs)

        ###
        mm_size = display_device.get('mm_size')
        if mm_size:
            mm_size = mm_size['width'], mm_size['height'],
        pixel_res = display_device.get('pixel_res')
        eye_distance = display_device.get('eye_distance')
        self.visual_angle_calc = VisualAngleCalc(mm_size,
                                                 pixel_res,
                                                 eye_distance)
        self.pix2deg = self.visual_angle_calc.pix2deg

    @property
    def filter_id(self):
        return 23

    @property
    def input_event_types(self):
        event_type_and_filter_ids = dict()
        event_type_and_filter_ids[BINOCULAR_EYE_SAMPLE] = [0, ]
        event_type_and_filter_ids[MONOCULAR_EYE_SAMPLE] = [0, ]
        return event_type_and_filter_ids

    def process(self):
        """"""
        samples_for_processing = []
        for in_evt in self.getInputEvents():
            if self.sample_type is None:
                self.initializeForSampleType(in_evt)

            # If event is binocular, convert to monocular.
            # Regardless of type, convert pix to angle positions and calculate
            # unfiltered velocity data.
            current_mono_evt = self.convertEvent(self.last_sample, in_evt)

            is_valid = self.isValidSample(current_mono_evt)
            if is_valid:
                # If sample is valid (no missing pos data), first
                # check for a previous missing data run and handle.
                if self.invalid_samples_run:
                    if self.last_valid_sample:
                        samples_for_processing.extend(
                            self.interpolateMissingData(current_mono_evt))
                        self._addVelocity(
                            samples_for_processing[-1], current_mono_evt)
                    # Discard all invalid samples that occurred prior
                    # to the first valid sample.
                    del self.invalid_samples_run[:]

                # Then add current event to field filters. If a filtered event
                # is returned, add it to the to be processed sample list.
                filtered_event = self.addToFieldFilters(current_mono_evt)
                if filtered_event:
                    filtered_event, _junk = filtered_event
                    x_vel_thresh, y_vel_thresh = self.addVelocityToAdaptiveThreshold(
                        filtered_event)
                    filtered_event[self.io_event_ix('raw_x')] = x_vel_thresh
                    filtered_event[self.io_event_ix('raw_y')] = y_vel_thresh
                    samples_for_processing.append(filtered_event)
                self.last_valid_sample = current_mono_evt
            else:
                self.invalid_samples_run.append(current_mono_evt)
                self.addOutputEvent(current_mono_evt)

            self.last_sample = current_mono_evt

        # Add any new filtered samples to be output.
        # Also create parsed events with no heuristics being used
        # at this point.
        for s in samples_for_processing:
            self.parseEvent(s)
            if self.isValidSample(s):
                self.addOutputEvent(s)

        self.clearInputEvents()

    def parseEvent(self, sample):
        if self._last_parser_sample:
            last_sec = self.getSampleEventCategory(self._last_parser_sample)
            current_sec = self.getSampleEventCategory(sample)
            if last_sec and last_sec != current_sec:
                start_event, end_event = self.createEyeEvents(
                    last_sec, current_sec, self._last_parser_sample, sample)
                if start_event:
                    self.addOutputEvent(start_event)
                if end_event:
                    self.addOutputEvent(end_event)
            else:
                self.open_parser_events.setdefault(
                    current_sec + '_SAMPLES', []).append(sample)
        self._last_parser_sample = sample

    def getSampleEventCategory(self, sample):
        if self.isValidSample(sample):
            x_velocity_threshold = sample[self.io_event_ix('raw_x')]
            y_velocity_threshold = sample[self.io_event_ix('raw_y')]
            if x_velocity_threshold == np.NaN:
                return None
            sample_vx = sample[self.io_event_ix('velocity_x')]
            sample_vy = sample[self.io_event_ix('velocity_y')]
            if sample_vx >= x_velocity_threshold or sample_vy >= y_velocity_threshold:
                return 'SAC'
            return 'FIX'
        return 'MIS'

    def createEyeEvents(
            self,
            last_sample_category,
            current_sample_category,
            last_sample,
            current_sample):
        start_event = None
        end_event = None

        if last_sample_category == 'MIS':
            # Create end blink event
            existing_start_event = self.open_parser_events.get('MIS')
            evt_samples = self.open_parser_events.get('MIS_SAMPLES')
            if evt_samples:
                del self.open_parser_events['MIS_SAMPLES']

            if existing_start_event:
                end_event = self.createBlinkEndEventArray(
                    last_sample, existing_start_event, evt_samples)
                del self.open_parser_events['MIS']
            else:
                # print2err("PARSER Warning: Blink Start Event not found; Blink End event being dropped: ", end_event)
                pass
        elif last_sample_category == 'FIX':
            # Create end fix event
            existing_start_event = self.open_parser_events.get('FIX')
            evt_samples = self.open_parser_events.get('FIX_SAMPLES')
            if evt_samples:
                del self.open_parser_events['FIX_SAMPLES']
            if existing_start_event:
                end_event = self.createFixationEndEventArray(
                    last_sample, existing_start_event, evt_samples)
                del self.open_parser_events['FIX']
            else:
                # print2err("PARSER Warning: Fixation Start Event not found; Fixation End event being dropped: ", end_event)
                pass
        elif last_sample_category == 'SAC':
            # Create end sac event
            existing_start_event = self.open_parser_events.get('SAC')
            evt_samples = self.open_parser_events.get('SAC_SAMPLES')
            if evt_samples:
                del self.open_parser_events['SAC_SAMPLES']
            if existing_start_event:
                end_event = self.createSaccadeEndEventArray(
                    last_sample, existing_start_event, evt_samples)
                del self.open_parser_events['SAC']
            else:
                # print2err("PARSER Warning: Saccade Start Event not found; Saccade End event being dropped: ", end_event)
                pass

        if current_sample_category == 'MIS':
            # Create start blink event
            start_event = self.createBlinkStartEventArray(current_sample)
            self.open_parser_events['MIS_SAMPLES'] = [current_sample, ]
            existing_start_event = self.open_parser_events.get('MIS')
            if existing_start_event:
                print2err(
                    'PARSER ERROR: Blink Start Event already Open and is being dropped: ',
                    existing_start_event)
            self.open_parser_events['MIS'] = current_sample

        elif current_sample_category == 'FIX':
            # Create start fix event
            start_event = self.createFixationStartEventArray(current_sample)
            self.open_parser_events['FIX_SAMPLES'] = [current_sample, ]
            existing_start_event = self.open_parser_events.get('FIX')
            if existing_start_event:
                print2err(
                    'PARSER ERROR: Fixation Start Event already Open and is being dropped: ',
                    existing_start_event)
            self.open_parser_events['FIX'] = current_sample

        elif current_sample_category == 'SAC':
            # Create start sac event
            start_event = self.createSaccadeStartEventArray(current_sample)
            self.open_parser_events['SAC_SAMPLES'] = [current_sample, ]
            existing_start_event = self.open_parser_events.get('SAC')
            if existing_start_event:
                print2err(
                    'PARSER ERROR: Saccade Start Event already Open and is being dropped: ',
                    existing_start_event)
            self.open_parser_events['SAC'] = current_sample

        return end_event, start_event

    def addVelocityToAdaptiveThreshold(self, sample):
        velocity_x = sample[self.io_event_ix('velocity_x')]
        velocity_y = sample[self.io_event_ix('velocity_y')]
        velocity_buffers = [
            self.adaptive_x_vthresh_buffer,
            self.adaptive_y_vthresh_buffer]
        velocity_buffer_indexs = [
            self.x_vthresh_buffer_index,
            self.y_vthresh_buffer_index]
        vthresh_values = []
        for v, velocity in enumerate([velocity_x, velocity_y]):
            current_velocity_buffer = velocity_buffers[v]
            current_vbuffer_index = velocity_buffer_indexs[v]
            blen = len(current_velocity_buffer)
            if velocity > 0.0:
                i = current_vbuffer_index % blen
                current_velocity_buffer[i] = velocity
                full = current_vbuffer_index >= blen
                if v == 0:
                    self.x_vthresh_buffer_index += 1
                else:
                    self.y_vthresh_buffer_index += 1
                if full:
                    PT = current_velocity_buffer.min() + current_velocity_buffer.std() * 3.0
                    velocity_below_thresh = current_velocity_buffer[
                        current_velocity_buffer < PT]
                    PTd = 2.0
                    pt_list = [PT, ]
                    while PTd >= 1.0:
                        if len(pt_list) > 0:
                            PT = velocity_below_thresh.mean() + 3.0 * velocity_below_thresh.std()
                            velocity_below_thresh = current_velocity_buffer[
                                current_velocity_buffer < PT]
                            PTd = np.abs(PT - pt_list[-1])
                        pt_list.append(PT)
                    vthresh_values.append(PT)
            if len(vthresh_values) != v + 1:
                vthresh_values.append(np.NaN)
        return vthresh_values

    def reset(self):
        eventfilters.DeviceEventFilter.reset(self)
        self._last_parser_sample = None
        self.last_valid_sample = None
        self.last_sample = None
        self.invalid_samples_run = []
        self.open_parser_events.clear()
        self.x_position_filter.clear()
        self.y_position_filter.clear()
        self.x_velocity_filter.clear()
        self.y_velocity_filter.clear()
        self.xy_velocity_filter.clear()
        self.x_vthresh_buffer_index = 0
        self.y_vthresh_buffer_index = 0

    def initializeForSampleType(self, in_evt):
        # in_evt[DeviceEvent.EVENT_TYPE_ID_INDEX]
        self.sample_type = MONOCULAR_EYE_SAMPLE
        #print2err("self.sample_type: ",self.sample_type,", ",EventConstants.getName(self.sample_type))
        self.io_sample_class = EventConstants.getClass(self.sample_type)
        self.io_event_fields = self.io_sample_class.CLASS_ATTRIBUTE_NAMES
        #print2err("self.io_sample_class: ",self.io_sample_class,", ",len(self.io_event_fields),"\n>>",self.io_event_fields)
        self.io_event_ix = self.io_sample_class.CLASS_ATTRIBUTE_NAMES.index

        if in_evt[DeviceEvent.EVENT_TYPE_ID_INDEX] == BINOCULAR_EYE_SAMPLE:
            self.convertEvent = self._convertToMonoAveraged
            self.isValidSample = lambda x: x[self.io_event_ix('status')] != 22
        else:
            self.convertEvent = self._convertMonoFields
            self.isValidSample = lambda x: x[self.io_event_ix('status')] == 0

    def interpolateMissingData(self, current_sample):
        samples_for_processing = []
        invalid_sample_count = len(self.invalid_samples_run)
        gx_ix = self.io_event_ix('angle_x')
        gy_ix = self.io_event_ix('angle_y')
        ps_ix = self.io_event_ix('pupil_measure1')
        starting_gx = self.last_valid_sample[gx_ix]
        starting_gy = self.last_valid_sample[gy_ix]
        starting_ps = self.last_valid_sample[ps_ix]
        ending_gx = current_sample[gx_ix]
        ending_gy = current_sample[gy_ix]
        ending_ps = current_sample[ps_ix]
        x_interp = np.linspace(starting_gx, ending_gx,
                               num=invalid_sample_count + 2)[1:-1]
        y_interp = np.linspace(starting_gy, ending_gy,
                               num=invalid_sample_count + 2)[1:-1]
        p_interp = np.linspace(starting_ps, ending_ps,
                               num=invalid_sample_count + 2)[1:-1]
#       print2err('>>>>')
#        print2err('invalid_sample_count: ', invalid_sample_count)
#        print2err('starting_gx, ending_gx: ', starting_gx,', ',ending_gx)
#        print2err('x_interp: ', x_interp)
#        print2err('starting_gy, ending_gy: ', starting_gx,', ',ending_gx)
#        print2err('y_interp: ', y_interp)
#        print2err('<<<<')

        prev_samp = self.last_valid_sample
        # interpolate missing sample values, adding to pos and vel filters
        for ix, curr_samp in enumerate(self.invalid_samples_run):
            curr_samp[gx_ix] = x_interp[ix]
            curr_samp[gy_ix] = y_interp[ix]
            curr_samp[ps_ix] = p_interp[ix]
            self._addVelocity(prev_samp, curr_samp)
            filtered_event = self.addToFieldFilters(curr_samp)
            if filtered_event:
                filtered_event, _junk = filtered_event
                samples_for_processing.append(filtered_event)
            prev_samp = curr_samp
        return samples_for_processing

    def addToFieldFilters(self, sample):
        self.x_position_filter.add(sample)
        self.y_position_filter.add(sample)
        self.x_velocity_filter.add(sample)
        self.y_velocity_filter.add(sample)
        return self.xy_velocity_filter.add(sample)

    def _convertPosToAngles(self, mono_event):
        gx_ix = self.io_event_ix('gaze_x')
        gx_iy = self.io_event_ix('gaze_y')
        mono_event[
            self.io_event_ix('angle_x')], mono_event[
            self.io_event_ix('angle_y')] = self.pix2deg(
            mono_event[gx_ix], mono_event[gx_iy])

    def _addVelocity(self, prev_event, current_event):
        io_ix = self.io_event_ix

        dx = np.abs(
            current_event[
                io_ix('angle_x')] -
            prev_event[
                io_ix('angle_x')])
        dy = np.abs(
            current_event[
                io_ix('angle_y')] -
            prev_event[
                io_ix('angle_y')])
        dt = current_event[io_ix('time')] - prev_event[io_ix('time')]

        current_event[io_ix('velocity_x')] = dx / dt
        current_event[io_ix('velocity_y')] = dy / dt
        current_event[io_ix('velocity_xy')] = np.hypot(dx / dt, dy / dt)

    def _convertMonoFields(self, prev_event, current_event):
        if self.isValidSample(current_event):
            self._convertPosToAngles(self, current_event)
            if prev_event:
                self._addVelocity(prev_event, current_event)

    def _convertToMonoAveraged(self, prev_event, current_event):
        mono_evt = []
        binoc_field_names = EventConstants.getClass(
            EventConstants.BINOCULAR_EYE_SAMPLE).CLASS_ATTRIBUTE_NAMES
        #print2err("binoc_field_names: ",len(binoc_field_names),"\n",binoc_field_names)
        status = current_event[binoc_field_names.index('status')]
        for field in self.io_event_fields:
            if field in binoc_field_names:
                mono_evt.append(current_event[binoc_field_names.index(field)])
            elif field == 'eye':
                mono_evt.append(LEFT_EYE)
            elif field.endswith('_type'):
                mono_evt.append(
                    int(current_event[binoc_field_names.index('left_%s' % (field))]))
            else:
                #print2err("binoc status: ",status)
                if status == 0:
                    lfv = float(
                        current_event[
                            binoc_field_names.index(
                                'left_%s' %
                                (field))])
                    rfv = float(
                        current_event[
                            binoc_field_names.index(
                                'right_%s' %
                                (field))])
                    mono_evt.append((lfv + rfv) / 2.0)
                elif status == 2:
                    mono_evt.append(
                        float(
                            current_event[
                                binoc_field_names.index(
                                    'left_%s' %
                                    (field))]))
                elif status == 20:
                    mono_evt.append(
                        float(
                            current_event[
                                binoc_field_names.index(
                                    'right_%s' %
                                    (field))]))
                elif status == 22:
                    # both eyes have missing data, so use data from left eye
                    # (does not really matter)
                    mono_evt.append(
                        float(
                            current_event[
                                binoc_field_names.index(
                                    'left_%s' %
                                    (field))]))
                else:
                    ValueError('Unknown Sample Status: %d' % (status))
        mono_evt[self.io_event_fields.index(
            'type')] = EventConstants.MONOCULAR_EYE_SAMPLE
        if self.isValidSample(mono_evt):
            self._convertPosToAngles(mono_evt)
            if prev_event:
                self._addVelocity(prev_event, mono_evt)
        return mono_evt

    def _binocSampleValidEyeData(self, sample):
        evt_status = sample[self.io_event_ix('status')]
        if evt_status == 0:
            # both eyes are valid
            return BOTH_EYE
        elif evt_status == 20:  # right eye data only
            return RIGHT_EYE
        elif evt_status == 2:  # left eye data only
            return LEFT_EYE
        elif evt_status == 22:  # both eye data missing
            return NO_EYE

    def createFixationStartEventArray(self, sample):
        return [sample[self.io_event_ix('experiment_id')],
                sample[self.io_event_ix('session_id')],
                sample[self.io_event_ix('device_id')],
                sample[self.io_event_ix('event_id')],
                EventConstants.FIXATION_START,
                sample[self.io_event_ix('device_time')],
                sample[self.io_event_ix('logged_time')],
                sample[self.io_event_ix('time')],
                0.0,
                0.0,
                0,
                sample[self.io_event_ix('eye')],
                sample[self.io_event_ix('gaze_x')],
                sample[self.io_event_ix('gaze_y')],
                0.0,
                sample[self.io_event_ix('angle_x')],
                sample[self.io_event_ix('angle_y')],
                # used to hold online x velocity threshold calculated for
                # sample
                sample[self.io_event_ix('raw_x')],
                # used to hold online y velocity threshold calculated for
                # sample
                sample[self.io_event_ix('raw_y')],
                sample[self.io_event_ix('pupil_measure1')],
                sample[self.io_event_ix('pupil_measure1_type')],
                0.0,
                0,
                0.0,
                0.0,
                sample[self.io_event_ix('velocity_x')],
                sample[self.io_event_ix('velocity_y')],
                sample[self.io_event_ix('velocity_xy')],
                sample[self.io_event_ix('status')]
                ]

    def createFixationEndEventArray(
            self,
            sample,
            existing_start_event,
            event_samples):
        evt_sample_array = np.asarray(event_samples)
        vx = self.io_event_ix('velocity_x')
        vy = self.io_event_ix('velocity_y')
        vxy = self.io_event_ix('velocity_xy')
        gx = self.io_event_ix('gaze_x')
        gy = self.io_event_ix('gaze_y')
        return [sample[self.io_event_ix('experiment_id')],
                sample[self.io_event_ix('session_id')],
                sample[self.io_event_ix('device_id')],
                sample[self.io_event_ix('event_id')],
                EventConstants.FIXATION_END,
                sample[self.io_event_ix('device_time')],
                sample[self.io_event_ix('logged_time')],
                sample[self.io_event_ix('time')],
                0.0,
                0.0,
                0,
                sample[self.io_event_ix('eye')],
                sample[self.io_event_ix(
                    'time')] - existing_start_event[self.io_event_ix('time')],
                existing_start_event[gx],
                existing_start_event[gy],
                0.0,
                existing_start_event[self.io_event_ix('angle_x')],
                existing_start_event[self.io_event_ix('angle_y')],
                # used to hold online x velocity threshold calculated for
                # sample
                existing_start_event[self.io_event_ix('raw_x')],
                # used to hold online y velocity threshold calculated for
                # sample
                existing_start_event[self.io_event_ix('raw_y')],
                existing_start_event[self.io_event_ix('pupil_measure1')],
                existing_start_event[self.io_event_ix('pupil_measure1_type')],
                0.0,
                0,
                0.0,
                0.0,
                existing_start_event[vx],
                existing_start_event[vy],
                existing_start_event[vxy],
                sample[gx],
                sample[gy],
                0.0,
                sample[self.io_event_ix('angle_x')],
                sample[self.io_event_ix('angle_y')],
                # used to hold online x velocity threshold calculated for
                # sample
                sample[self.io_event_ix('raw_x')],
                # used to hold online y velocity threshold calculated for
                # sample
                sample[self.io_event_ix('raw_y')],
                sample[self.io_event_ix('pupil_measure1')],
                sample[self.io_event_ix('pupil_measure1_type')],
                0.0,
                0,
                0.0,
                0.0,
                sample[vx],
                sample[vy],
                sample[vxy],
                evt_sample_array[:, gx].mean(),  # average_gaze_x,
                evt_sample_array[:, gy].mean(),  # average_gaze_y,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                evt_sample_array[
            :,
            self.io_event_ix('pupil_measure1')].mean(),
            # average_pupil_measure1,
            # average_pupil_measure1_type,
            sample[self.io_event_ix('pupil_measure1_type')],
            0.0,
            0.0,
            0.0,
            0.0,
            evt_sample_array[:, vx].mean(),  # average_velocity_x,
            evt_sample_array[:, vy].mean(),  # average_velocity_y,
            evt_sample_array[:, vxy].mean(),  # average_velocity_xy,
            evt_sample_array[:, vx].max(),  # peak_velocity_x,
            evt_sample_array[:, vy].max(),  # peak_velocity_y,
            evt_sample_array[:, vxy].max(),  # peak_velocity_xy,
            sample[self.io_event_ix('status')]
        ]

    ################### Saccade Event Types ##########################

    def createSaccadeStartEventArray(self, sample):
        return [sample[self.io_event_ix('experiment_id')],
                sample[self.io_event_ix('session_id')],
                sample[self.io_event_ix('device_id')],
                sample[self.io_event_ix('event_id')],
                EventConstants.SACCADE_START,
                sample[self.io_event_ix('device_time')],
                sample[self.io_event_ix('logged_time')],
                sample[self.io_event_ix('time')],
                0.0,
                0.0,
                0,
                sample[self.io_event_ix('eye')],
                sample[self.io_event_ix('gaze_x')],
                sample[self.io_event_ix('gaze_y')],
                0.0,
                sample[self.io_event_ix('angle_x')],
                sample[self.io_event_ix('angle_y')],
                # used to hold online x velocity threshold calculated for
                # sample
                sample[self.io_event_ix('raw_x')],
                # used to hold online y velocity threshold calculated for
                # sample
                sample[self.io_event_ix('raw_y')],
                sample[self.io_event_ix('pupil_measure1')],
                sample[self.io_event_ix('pupil_measure1_type')],
                0.0,
                0,
                0.0,
                0.0,
                sample[self.io_event_ix('velocity_x')],
                sample[self.io_event_ix('velocity_y')],
                sample[self.io_event_ix('velocity_xy')],
                sample[self.io_event_ix('status')]
                ]

    def createSaccadeEndEventArray(
            self,
            sample,
            existing_start_event,
            event_samples):
        evt_sample_array = np.asarray(event_samples)
        gx = self.io_event_ix('gaze_x')
        gy = self.io_event_ix('gaze_y')
        x1 = existing_start_event[gx]
        y1 = existing_start_event[gy]
        x2 = sample[gx]
        y2 = sample[gy]
        xDiff = x2 - x1
        yDiff = y2 - y1
        vx = self.io_event_ix('velocity_x')
        vy = self.io_event_ix('velocity_y')
        vxy = self.io_event_ix('velocity_xy')
        return [sample[self.io_event_ix('experiment_id')],
                sample[self.io_event_ix('session_id')],
                sample[self.io_event_ix('device_id')],
                sample[self.io_event_ix('event_id')],
                EventConstants.SACCADE_END,
                sample[self.io_event_ix('device_time')],
                sample[self.io_event_ix('logged_time')],
                sample[self.io_event_ix('time')],
                0.0,
                0.0,
                0,
                sample[self.io_event_ix('eye')],
                sample[self.io_event_ix(
                    'time')] - existing_start_event[self.io_event_ix('time')],
                xDiff,
                yDiff,
                np.rad2deg(np.arctan(yDiff, xDiff)),
                existing_start_event[gx],
                existing_start_event[gy],
                0.0,
                existing_start_event[self.io_event_ix('angle_x')],
                existing_start_event[self.io_event_ix('angle_y')],
                # used to hold online x velocity threshold calculated for
                # sample
                existing_start_event[self.io_event_ix('raw_x')],
                # used to hold online y velocity threshold calculated for
                # sample
                existing_start_event[self.io_event_ix('raw_y')],
                existing_start_event[self.io_event_ix('pupil_measure1')],
                existing_start_event[self.io_event_ix('pupil_measure1_type')],
                0.0,
                0,
                0.0,
                0.0,
                existing_start_event[vx],
                existing_start_event[vy],
                existing_start_event[vxy],
                sample[gx],
                sample[gy],
                0.0,
                sample[self.io_event_ix('angle_x')],
                sample[self.io_event_ix('angle_y')],
                # used to hold online x velocity threshold calculated for
                # sample
                sample[self.io_event_ix('raw_x')],
                # used to hold online y velocity threshold calculated for
                # sample
                sample[self.io_event_ix('raw_y')],
                sample[self.io_event_ix('pupil_measure1')],
                sample[self.io_event_ix('pupil_measure1_type')],
                0.0,
                0,
                0.0,
                0.0,
                sample[vx],
                sample[vy],
                sample[vxy],
                evt_sample_array[:, vx].mean(),  # average_velocity_x,
                evt_sample_array[:, vy].mean(),  # average_velocity_y,
                evt_sample_array[:, vxy].mean(),  # average_velocity_xy,
                evt_sample_array[:, vx].max(),  # peak_velocity_x,
                evt_sample_array[:, vy].max(),  # peak_velocity_y,
                evt_sample_array[:, vxy].max(),  # peak_velocity_xy,
                sample[self.io_event_ix('status')]
                ]

    ################### Blink Event Types ##########################

    def createBlinkStartEventArray(self, sample):
        return [sample[self.io_event_ix('experiment_id')],
                sample[self.io_event_ix('session_id')],
                sample[self.io_event_ix('device_id')],
                sample[self.io_event_ix('event_id')],
                EventConstants.BLINK_START,
                sample[self.io_event_ix('device_time')],
                sample[self.io_event_ix('logged_time')],
                sample[self.io_event_ix('time')],
                0.0,
                0.0,
                0,
                sample[self.io_event_ix('eye')],
                sample[self.io_event_ix('status')]
                ]

    def createBlinkEndEventArray(
            self,
            sample,
            existing_start_event,
            event_samples):
        return [
            sample[
                self.io_event_ix('experiment_id')], sample[
                self.io_event_ix('session_id')], sample[
                self.io_event_ix('device_id')], sample[
                    self.io_event_ix('event_id')], EventConstants.BLINK_END, sample[
                        self.io_event_ix('device_time')], sample[
                            self.io_event_ix('logged_time')], sample[
                                self.io_event_ix('time')], 0.0, 0.0, 0, sample[
                                    self.io_event_ix('eye')], sample[
                                        self.io_event_ix('time')] - existing_start_event[
                                            self.io_event_ix('time')], sample[
                                                self.io_event_ix('status')]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from .project import DetailsPanel
from .functions import logInPavlovia

from ._base import BaseFrame
from psychopy.localization import _translate
from psychopy.projects import pavlovia

import copy
import wx
from pkg_resources import parse_version
from wx.lib import scrolledpanel as scrlpanel
import wx.lib.mixins.listctrl as listmixin

import requests

from .. import utils
from ..themes._themes import IconCache

_starred = u"\u2605"
_unstarred = u"\u2606"
_fork = u"\u2442"


class ListCtrl(wx.ListCtrl, listmixin.ListCtrlAutoWidthMixin):
    """
    A ListCtrl with ListCtrlAutoWidthMixin already mixed in, purely for convenience
    """
    def __init__(self, *args, **kwargs):
        wx.ListCtrl.__init__(self, *args, **kwargs)
        listmixin.ListCtrlAutoWidthMixin.__init__(self)


class SearchFrame(wx.Dialog):

    def __init__(self, app, parent=None, style=wx.RESIZE_BORDER | wx.DEFAULT_DIALOG_STYLE | wx.CENTER | wx.TAB_TRAVERSAL | wx.NO_BORDER,
                 pos=wx.DefaultPosition):
        self.frameType = 'ProjectSearch'
        wx.Dialog.__init__(self, parent, -1, title=_translate("Search for projects online"),
                           style=style,
                           size=(1080, 720), pos=pos)
        self.app = app
        self.SetMinSize((980, 520))
        # Setup sizer
        self.sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(self.sizer)
        # Make panels
        self.detailsPanel = DetailsPanel(self)
        self.searchPanel = SearchPanel(self, viewer=self.detailsPanel)
        # Add to sizers
        self.sizer.Add(self.searchPanel, border=12, flag=wx.EXPAND | wx.ALL)
        self.sizer.Add(self.detailsPanel, proportion=1, border=12, flag=wx.EXPAND | wx.ALL)
        # Layout
        self.Layout()


class SearchPanel(wx.Panel):
    """A scrollable panel showing a list of projects. To be used within the
    Project Search dialog
    """
    class FilterLabel(wx.StaticText):
        """
        A label to show what filters and sorting are applied to the current search
        """
        def update(self):
            # Get parent object
            parent = self.GetParent()
            # Add me mode
            mineLbl = ""
            if parent._mine:
                mineLbl += "Editable by me. "
            # Add sort params
            sortLbl = ""
            if len(parent.sortOrder):
                sortLbl += "Sort by: "
                sortLbl += " then ".join([item.label.lower() for item in parent.sortOrder])
                sortLbl += ". "
            # Add filter params
            filterLbl = ""
            for key, values in parent.filterTerms.items():
                if values:
                    if isinstance(values, str):
                        values = [values]
                    filterLbl += f"{key}: "
                    filterLbl += " or ".join(values)
                    filterLbl += ". "
            # Construct full label
            label = mineLbl + sortLbl + filterLbl
            # Apply label
            self.SetLabel(label)
            # Show or hide self according to label
            self.Show(bool(label))
            parent.Layout()

        def hoverOn(self, evt=None):
            self.Wrap(self.GetParent().GetSize()[0] - 12)
            self.GetParent().Layout()

        def hoverOff(self, evt=None):
            self.SetLabel(self.GetLabel().replace("\n", " "))
            self.GetParent().Layout()

    def __init__(self, parent, viewer,
                 size=(400, -1),
                 style=wx.NO_BORDER
                 ):
        # Init self
        wx.Panel.__init__(self, parent, -1,
                          size=size,
                          style=style
                          )
        iconCache = parent.app.iconCache
        self.session = pavlovia.getCurrentSession()
        # Setup sizer
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)
        # Add search ctrl
        self.searchCtrl = wx.SearchCtrl(self)
        self.searchCtrl.Bind(wx.EVT_SEARCH, self.search)
        self.sizer.Add(self.searchCtrl, border=4, flag=wx.EXPAND | wx.ALL)
        # Add button sizer
        self.btnSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.Add(self.btnSizer, border=4, flag=wx.EXPAND | wx.ALL)
        # Add "me mode" button
        self.mineBtn = wx.ToggleButton(self, size=(64, -1), label=_translate("Me"))
        self.mineBtn.SetBitmap(iconCache.getBitmap("person_off", size=16))
        self.mineBtn.SetBitmapFocus(iconCache.getBitmap("person_on", size=16))
        self.mineBtn.SetBitmapPressed(iconCache.getBitmap("person_on", size=16))
        self._mine = False
        self.mineBtn.Bind(wx.EVT_TOGGLEBUTTON, self.onMineBtn)
        self.mineBtn.Enable(self.session.userID is not None)
        self.btnSizer.Add(self.mineBtn, border=6, flag=wx.EXPAND | wx.RIGHT | wx.TOP | wx.BOTTOM)
        self.btnSizer.AddStretchSpacer(1)
        # Add sort button
        self.sortBtn = wx.Button(self, label=_translate("Sort..."))
        self.sortOrder = []
        self.sortBtn.Bind(wx.EVT_BUTTON, self.sort)
        self.btnSizer.Add(self.sortBtn, border=6, flag=wx.LEFT | wx.RIGHT | wx.TOP | wx.BOTTOM)
        # Add filter button
        self.filterBtn = wx.Button(self, label=_translate("Filter..."))
        self.filterTerms = {
            "Status": [],
            "Platform": [],
            "Keywords": [],
        }
        self.filterOptions = {
            "Author": None,
            "Status": ["running", "piloting", "inactive"],
            "Platform": ["psychojs", "jspsych", "labjs", "opensesame", "other"],
            "Visibility": ["public", "private"],
            "Keywords": None,
        }
        self.filterBtn.Bind(wx.EVT_BUTTON, self.filter)
        self.btnSizer.Add(self.filterBtn, border=6, flag=wx.LEFT | wx.TOP | wx.BOTTOM)
        # Add filter label
        self.filterLbl = self.FilterLabel(self, style=wx.ST_ELLIPSIZE_END)
        self.filterLbl.Bind(wx.EVT_ENTER_WINDOW, self.filterLbl.hoverOn)
        self.filterLbl.Bind(wx.EVT_LEAVE_WINDOW, self.filterLbl.hoverOff)
        self.sizer.Add(self.filterLbl, border=6, flag=wx.LEFT | wx.RIGHT)
        self.filterLbl.update()
        # Add project list
        self.projectList = ListCtrl(self, size=(-1, -1), style=wx.LC_REPORT | wx.LC_SINGLE_SEL)
        self.projectList.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.showProject)
        self.sizer.Add(self.projectList, proportion=1, border=4, flag=wx.EXPAND | wx.ALL)
        # Setup project list
        self.projectList.InsertColumn(0, _starred, width=36, format=wx.LIST_FORMAT_CENTER)  # Stars
        #self.projectList.InsertColumn(1, _translate('Status'), width=wx.LIST_AUTOSIZE, format=wx.LIST_FORMAT_LEFT)  # Status
        self.projectList.InsertColumn(2, _translate('Author'), width=wx.LIST_AUTOSIZE, format=wx.LIST_FORMAT_LEFT)  # Author
        self.projectList.InsertColumn(3, _translate('Name'), width=wx.LIST_AUTOSIZE, format=wx.LIST_FORMAT_LEFT)  # Name
        # Setup projects dict
        self.projects = None

        # Link ProjectViewer
        self.viewer = viewer
        self.projectList.Bind(wx.EVT_LIST_ITEM_SELECTED, self.showProject)
        # Layout
        self.Layout()

        # Initial search
        self.search()

    def search(self, evt=None):
        # Get search term
        if evt is None:
            term = self.searchCtrl.GetValue()
        elif isinstance(evt, str):
            term = evt
        else:
            term = evt.GetString()
        # Do search
        self.projects = pavlovia.PavloviaSearch(term=term, filterBy=self.filterTerms, mine=self._mine)

        # Sort values
        if len(self.projects):
            self.projects.sort_values(self.sortOrder)

        # Refresh
        self.refreshCtrl()

    def refreshCtrl(self):
        # Clear list and projects dict
        self.projectList.DeleteAllItems()
        # Skip if not searched yet
        if self.projects is None:
            return
        # Skip if empty search
        if len(self.projects) == 0:
            return
        # Populate list and projects dict
        for i, _ in self.projects.iterrows():
            i = self.projectList.Append([
                self.projects['nbStars'][i],
                #self.projects['status'][i],
                self.projects['pathWithNamespace'][i].split('/')[0],
                self.projects['name'][i],
            ])

    def sort(self, evt=None):
        # Create temporary arrays
        items = copy.deepcopy(self.sortOrder)
        selected = []
        # Append missing items to copy
        for item in pavlovia.PavloviaSearch.sortTerms:
            if item in self.sortOrder:
                selected.append(True)
            else:
                items.append(copy.deepcopy(item))
                selected.append(False)
        # Create dlg
        _dlg = SortDlg(self, items=items, selected=selected)
        if _dlg.ShowModal() != wx.ID_OK:
            return
        # Update sort order
        self.sortOrder = _dlg.ctrls.GetValue()
        # Sort
        if self.projects is not None:
            self.projects.sort_values(self.sortOrder)
        # Refresh
        self.refreshCtrl()
        self.filterLbl.update()

    def filter(self, evt=None):
        # Open filter dlg
        _dlg = FilterDlg(self,
                         terms=self.filterTerms,
                         options=self.filterOptions)
        # Skip if cancelled
        if _dlg.ShowModal() != wx.ID_OK:
            return
        # Update filters if Okayed
        self.filterTerms = _dlg.GetValue()
        # Update filter label
        self.filterLbl.update()
        # Re-search
        self.search()

    def showProject(self, evt=None):
        """
        View current project in associated viewer
        """
        if self.projects is not None:
            proj = self.projects.iloc[self.projectList.GetFocusedItem()]
            # Set project to None while waiting
            self.viewer.project = None
            # Set project
            self.viewer.project = pavlovia.PavloviaProject(proj)

    def onMineBtn(self, evt=None):
        # If triggered manually with a bool, do setting
        if isinstance(evt, bool):
            self.mineBtn.Value = evt
        # Apply "mine" filter
        self._mine = self.mineBtn.Value
        # Clear and disable filters & search todo: This is a stop gap, remove once the Pavlovia API can accept searches and filters WITHIN a designer
        self.filterBtn.Enable(not self._mine)
        self.searchCtrl.Enable(not self._mine)
        if self._mine:
            self.filterTerms = {key: [] for key in self.filterTerms}
            self.searchCtrl.Value = ""
        # Update
        self.filterLbl.update()
        self.search()


class SortDlg(wx.Dialog):
    def __init__(self, parent, size=(200, 400),
                 items=pavlovia.PavloviaSearch.sortTerms,
                 selected=False):
        wx.Dialog.__init__(self, parent, size=size, title="Sort by...", style=wx.DEFAULT_DIALOG_STYLE | wx.DIALOG_NO_PARENT)
        # Setup sizer
        self.contentBox = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.contentBox)
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.contentBox.Add(self.sizer, proportion=1, border=0, flag=wx.EXPAND | wx.ALL)
        # Create rearrange control
        self.ctrls = utils.SortCtrl(self,
                                    items=items,
                                    showFlip=True,
                                    showSelect=True, selected=selected)
        self.sizer.Add(self.ctrls, border=6, flag=wx.EXPAND | wx.ALL)
        # Add Okay button
        self.sizer.AddStretchSpacer(1)
        self.OkayBtn = wx.Button(self, id=wx.ID_OK, label="Okay")
        self.contentBox.Add(self.OkayBtn, border=6, flag=wx.ALL | wx.ALIGN_RIGHT)
        # Bind cancel
        self.Bind(wx.EVT_CLOSE, self.onCancel)

    def onCancel(self, evt=None):
        self.EndModal(wx.ID_CANCEL)


class FilterDlg(wx.Dialog):
    class KeyCtrl(wx.Window):
        def __init__(self, parent,
                     key, value,
                     options=None,
                     selected=True):
            # Init self
            wx.Window.__init__(self, parent, size=(-1, -1))
            self.SetBackgroundColour(parent.GetBackgroundColour())
            # Create master sizer
            self.sizer = wx.BoxSizer(wx.VERTICAL)
            self.SetSizer(self.sizer)
            # Create title sizer
            self.titleSizer = wx.BoxSizer(wx.HORIZONTAL)
            self.sizer.Add(self.titleSizer, border=0, flag=wx.ALL | wx.EXPAND)
            # Add tickbox
            self.selectCtrl = wx.CheckBox(self)
            self.selectCtrl.Bind(wx.EVT_CHECKBOX, self.onSelect)
            self.selectCtrl.SetValue(selected)
            self.titleSizer.Add(self.selectCtrl, border=6, flag=wx.ALL | wx.EXPAND)
            # Add label
            self.key = key
            self.label = wx.StaticText(self, label=f"{key}:")
            self.titleSizer.Add(self.label, proportion=1, border=6, flag=wx.ALL | wx.EXPAND | wx.TEXT_ALIGNMENT_LEFT)
            # Add ctrl
            if options is None:
                self.ctrl = wx.TextCtrl(self, value=",".join(value))
            else:
                self.ctrl = wx.CheckListBox(self, choices=options)
                self.ctrl.SetCheckedStrings(value)
            self.sizer.Add(self.ctrl, border=6, flag=wx.LEFT | wx.RIGHT | wx.EXPAND)
            # Layout
            self.Layout()
            # Enable
            self.Enable(selected)

        @property
        def selected(self):
            return self.selectCtrl.Value

        @selected.setter
        def selected(self, value):
            self.selectCtrl.Value = value

        def onSelect(self, evt=None):
            self.Enable(self.selected)

        def Enable(self, enable=True):
            # Select button always enabled
            self.selectCtrl.Enable(True)
            # Enable/disable children
            self.label.Enable(enable)
            self.ctrl.Enable(enable)
            if not enable and isinstance(self.ctrl, wx.CheckListBox):
                # If disabled, every option should be checked and none selected
                self.ctrl.SetCheckedStrings(self.ctrl.GetStrings())
                self.ctrl.SetSelection(-1)

        def Disable(self):
            self.Enable(False)

        def GetValue(self):
            # If deselected, return blank
            if not self.selected:
                return []
            # Otherwise, get ctrl value
            if isinstance(self.ctrl, wx.CheckListBox):
                # List of checked strings for check lists
                return self.ctrl.GetCheckedStrings()
            else:
                # String split by commas for text ctrl
                if self.ctrl.GetValue():
                    return self.ctrl.GetValue().split(",")
                else:
                    # Substitute [''] for [] so it's booleanised to False
                    return []

    def __init__(self, parent, size=(250, 400),
                 terms={},
                 options={}):
        wx.Dialog.__init__(self, parent, size=size,
                           title="Filter by...", style=wx.DEFAULT_DIALOG_STYLE | wx.DIALOG_NO_PARENT)
        # Setup sizer
        self.contentBox = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.contentBox)
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.contentBox.Add(self.sizer, proportion=1, border=0, flag=wx.EXPAND | wx.ALL)
        # Add dict ctrl
        self.ctrls = {}
        for key in terms:
            self.ctrls[key] = self.KeyCtrl(self, key,
                                           terms[key], selected=bool(terms[key]),
                                           options=options[key])
            self.sizer.Add(self.ctrls[key], border=6, flag=wx.EXPAND | wx.ALL)
        # Add Okay button
        self.sizer.AddStretchSpacer(1)
        self.OkayBtn = wx.Button(self, id=wx.ID_OK, label=_translate("OK"))
        self.contentBox.Add(self.OkayBtn, border=6, flag=wx.ALL | wx.ALIGN_RIGHT)
        # Bind cancel
        self.Bind(wx.EVT_CLOSE, self.onCancel)

    def onCancel(self, evt=None):
        self.EndModal(wx.ID_CANCEL)

    def GetValue(self):
        # Create blank dict
        value = {}
        # Update dict keys with value from each ctrl
        for key, ctrl in self.ctrls.items():
            value[key] = ctrl.GetValue()

        return value


def sortProjects(seq, name, reverse=False):
    return sorted(seq, key=lambda k: k[name], reverse=reverse)


def getUniqueByID(seq):
    """Very fast function to remove duplicates from a list while preserving order

    Based on sort f8() by Dave Kirby
    benchmarked at https://www.peterbe.com/plog/uniqifiers-benchmark

    Requires Python>=2.7 (requires set())
    """
    # Order preserving
    seen = set()
    return [x for x in seq if x.id not in seen and not seen.add(x.id)]



import codecs
import shutil
import sys
from pathlib import Path
from tempfile import mkdtemp

import numpy as np

from ..utils import TESTS_DATA_PATH
from psychopy import experiment, core


class TestLoops:
    @classmethod
    def setup_class(cls):
        """
        Load and run various experiments just once and use the objects / output in later tests
        """
        # Setup temporary dir
        try:
            cls.tempDir = mkdtemp(dir=Path(__file__).root, prefix='psychopy-tests-loops')
        except (PermissionError, OSError):
            # can't write to root on Linux
            cls.tempDir = mkdtemp(prefix='psychopy-tests-loops')
        # List of filenames for experiments to run
        filenames = [
            'testLoopsBlocks',
            'testStaircase',
            'test_current_loop_attr'
        ]
        # Run each experiment to get data
        cls.cases = {}
        for filename in filenames:
            # Copy file to temp dir so it's in the same folder as we want data to output to
            ogExpFile = Path(TESTS_DATA_PATH) / "test_loops" / f"{filename}.psyexp"
            expFile = Path(cls.tempDir) / f"{filename}.psyexp"
            shutil.copy(ogExpFile, expFile)
            # Load experiment from file
            exp = experiment.Experiment()
            exp.loadFromXML(expFile)
            # Change data file output to temp dir
            datafile = (Path(cls.tempDir) / "data" / f"{filename}.csv")
            exp.settings.params['Data filename'].val = f"'data' + os.sep + '{filename}'"
            # Write scripts
            pyScript = exp.writeScript(target="PsychoPy")
            # jsScript = exp.writeScript(target="PsychoJS")  # disabled until all loops work in JS
            # Save Python script to temp dir
            pyScriptFile = Path(cls.tempDir) / f"{filename}.py"
            with codecs.open(str(pyScriptFile), 'w', 'utf-8-sig') as f:
                f.write(pyScript)

            # Run Python script to generate data file
            stdout, stderr = core.shellCall([sys.executable, str(pyScriptFile)], stderr=True)
            # Load data file
            with open(datafile, "rb") as f:
                data = np.recfromcsv(f, case_sensitive=True)

            # Store
            cls.cases[filename] = {
                'exp': exp,
                'pyScript': pyScript,
                # 'jsScript': jsScript,  # disabled until all loops work in JS
                'data': data,
                'stdout': stdout,
                'stderr': stderr,
            }

    def test_output_length(self):
        """
        Check that experiments with loops produce data of the correct length
        """
        # Define desired length for each case
        answers = {
            'testLoopsBlocks': 8,  # because 4 'blocks' with 2 trials each (3 stims per trial)
            'testStaircase': 5,  # 5 reps
        }
        # Test each case
        for filename, case in self.cases.items():
            if filename in answers:
                assert len(case['data']) == answers[filename], (
                    f"Expected array {answers[filename]} long, received:\n"
                    f"{case['data']}"
                )

    def test_current_loop_attr(self):
        assert "___FAILURE___" not in self.cases['test_current_loop_attr']['stderr'], (
            self.cases['test_current_loop_attr']['stderr']
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Errors and warnings associated with audio recording and playback.
"""

from ..exceptions import SoundFormatError, DependencyError

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

__all__ = [
    'AudioStreamError',
    'AudioFrequencyRangeError',
    'AudioUnsupportedSampleRateError',
    'AudioInvalidDeviceError',
    'AudioUnsupportedCodecError',
    'AudioInvalidCaptureDeviceError',
    'AudioInvalidPlaybackDeviceError',
    'AudioRecordingBufferFullError',
    'RecognizerAPICredentialsError',
    'RecognizerLanguageNotSupportedError',
    'RecognizerEngineNotFoundError',
    'SoundFormatError',
    'DependencyError'
]


# ------------------------------------------------------------------------------
# Audio hardware and software exceptions
#

class AudioUnsupportedCodecError(Exception):
    """Error raise when trying to save or load and unsupported audio
    codec/format.

    """
    pass


class AudioStreamError(Exception):
    """Error raised when there is a problem during audio recording/streaming."""
    pass


class AudioFrequencyRangeError(Exception):  # might transform to a warning
    """Error raised when generating a tone with a frequency outside the audible
    range for humans (20Hz to 20kHz).

    """
    pass


class AudioUnsupportedSampleRateError(Exception):
    """Error raised when the sampling rate is not supported by the hardware.

    """
    pass


class AudioInvalidDeviceError(Exception):
    """Error raised when the audio device configuration does not match any
    supported configuration.

    """
    pass


class AudioInvalidCaptureDeviceError(AudioInvalidDeviceError):
    """Error raised when the audio device cannot be used for capture (i.e. not
    a microphone).

    """
    pass


class AudioInvalidPlaybackDeviceError(AudioInvalidDeviceError):
    """Error raised when the audio device is not suitable for playback.

    """
    pass


class AudioRecordingBufferFullError(Exception):
    """Error raised when the recording buffer is full."""
    pass


# ------------------------------------------------------------------------------
# Transcriber exceptions
#

class RecognizerAPICredentialsError(ValueError):
    """Raised when a given speech recognition is being given improper
    credentials (i.e. API key is invalid or not found).
    """


class RecognizerLanguageNotSupportedError(ValueError):
    """Raised when the specified language is not supported by the engine. If you
    get this, you need to install the appropriate language support or select
    another language.
    """


class RecognizerEngineNotFoundError(ModuleNotFoundError):
    """Raised when the specified recognizer cannot be found. Usually get this
    error if the required packages are not installed for a recognizer that is
    invoked.
    """


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#pylint: skip-file

# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------

from pyglet.gl import (GLuint, glEnable, GL_TEXTURE_2D, glBindTexture, glTexParameteri,
                       GL_TEXTURE_WRAP_S, GL_CLAMP, GL_TEXTURE_WRAP_T, glTexImage2D,
                       GL_TEXTURE_MIN_FILTER, GL_LINEAR, GL_TEXTURE_MAG_FILTER, GL_ALPHA,
                       GL_UNSIGNED_BYTE, GL_RGB, GL_RGBA, glGenTextures)
import ctypes
import math
import numpy as np
import sys


class TextureAtlas:
    '''
    Group multiple small data regions into a larger texture.

    The algorithm is based on the article by Jukka Jylänki : "A Thousand Ways
    to Pack the Bin - A Practical Approach to Two-Dimensional Rectangle Bin
    Packing", February 27, 2010. More precisely, this is an implementation of
    the Skyline Bottom-Left algorithm based on C++ sources provided by Jukka
    Jylänki at: http://clb.demon.fi/files/RectangleBinPack/

    Example usage:
    --------------

    atlas = TextureAtlas(512,512,3)
    region = atlas.get_region(20,20)
    ...
    atlas.set_region(region, data)
    '''

    def __init__(self, width=1024, height=1024, depth=1):
        '''
        Initialize a new atlas of given size.

        Parameters
        ----------

        width : int
            Width of the underlying texture

        height : int
            Height of the underlying texture

        depth : 1 or 3
            Depth of the underlying texture
        '''
        super(TextureAtlas, self).__init__()
        self.width = int(math.pow(2, int(math.log(width, 2) + 0.5)))
        self.height = int(math.pow(2, int(math.log(height, 2) + 0.5)))
        self.depth = depth
        self.nodes = [(0, 0, self.width), ]
        self.data = np.zeros((self.height, self.width, self.depth),
                             dtype=np.ubyte)
        self.texid = None
        self.used = 0
        self.max_y = 0

    def getTextureID(self):
        return self.texid

    def upload(self):
        '''
        Upload atlas data into video memory.
        '''
        glEnable(GL_TEXTURE_2D)
        if self.texid is None:
            self.texid = GLuint(0)
            glGenTextures(1, ctypes.byref(self.texid))
        glBindTexture(GL_TEXTURE_2D, self.texid)
        glTexParameteri(GL_TEXTURE_2D,
                        GL_TEXTURE_WRAP_S, GL_CLAMP)
        glTexParameteri(GL_TEXTURE_2D,
                        GL_TEXTURE_WRAP_T, GL_CLAMP)
        glTexParameteri(GL_TEXTURE_2D,
                        GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D,
                        GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        if self.depth == 1:
            glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA,
                         self.width, self.height, 0,
                         GL_ALPHA, GL_UNSIGNED_BYTE, self.data.ctypes)
        elif self.depth == 3:
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB,
                         self.width, self.height, 0,
                         GL_RGB, GL_UNSIGNED_BYTE, self.data.ctypes)
        else:
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
                         self.width, self.height, 0,
                         GL_RGBA, GL_UNSIGNED_BYTE, self.data.ctypes)
        glBindTexture(GL_TEXTURE_2D, 0)

    def resize(self, new_height):
        # np.zeros((self.height, self.width, self.depth),
        self.data = self.data[:new_height]
 #                              dtype=np.ubyte)
        self.height = new_height

    def set_region(self, region, data):
        '''
        Set a given region width provided data.

        Parameters
        ----------

        region : (int,int,int,int)
            an allocated region (x,y,width,height)

        data : numpy array
            data to be copied into given region
        '''

        x, y, width, height = region
        self.data[y:y + height, x:x + width, :] = data

    def get_region(self, width, height):
        '''
        Get a free region of given size and allocate it

        Parameters
        ----------

        width : int
            Width of region to allocate

        height : int
            Height of region to allocate

        Return
        ------
            A newly allocated region as (x,y,width,height) or (-1,-1,0,0)
        '''

        best_height = sys.maxsize
        best_index = -1
        best_width = sys.maxsize
        region = 0, 0, width, height

        for i in range(len(self.nodes)):
            y = self.fit(i, width, height)
            if y >= 0:
                node = self.nodes[i]
                if (y + height < best_height or
                        (y + height == best_height and node[2] < best_width)):
                    best_height = y + height
                    best_index = i
                    best_width = node[2]
                    region = node[0], y, width, height

        if best_index == -1:
            return -1, -1, 0, 0

        node = region[0], region[1] + height, width
        self.nodes.insert(best_index, node)

        i = best_index + 1
        while i < len(self.nodes):
            node = self.nodes[i]
            prev_node = self.nodes[i - 1]
            if node[0] < prev_node[0] + prev_node[2]:
                shrink = prev_node[0] + prev_node[2] - node[0]
                x, y, w = self.nodes[i]
                self.nodes[i] = x + shrink, y, w - shrink
                if self.nodes[i][2] <= 0:
                    del self.nodes[i]
                    i -= 1
                else:
                    break
            else:
                break
            i += 1

        self.max_y = region[1] + region[3]
        self.merge()
        self.used += width * height
        return region

    def fit(self, index, width, height):
        '''
        Test if region (width,height) fit into self.nodes[index]

        Parameters
        ----------

        index : int
            Index of the internal node to be tested

        width : int
            Width or the region to be tested

        height : int
            Height or the region to be tested

        '''
        node = self.nodes[index]
        x, y = node[0], node[1]
        width_left = width

        if x + width > self.width:
            return -1

        i = index
        while width_left > 0:
            node = self.nodes[i]
            y = max(y, node[1])
            if y + height > self.height:
                return -1
            width_left -= node[2]
            i += 1
        return y

    def merge(self):
        '''
        Merge nodes
        '''
        i = 0
        while i < len(self.nodes) - 1:
            node = self.nodes[i]
            next_node = self.nodes[i + 1]
            if node[1] == next_node[1]:
                self.nodes[i] = node[0], node[1], node[2] + next_node[2]
                del self.nodes[i + 1]
            else:
                i += 1

    def totalArea(self):
        return self.width * self.height

    def usedArea(self):
        return self.used

    def freeArea(self):
        return self.totalArea() - self.usedArea()

    def __del__(self):
        self.texid = None
        self.data = None
        self.nodes = None
